"""Builder for Current and Pending Reports."""

import datetime as dt
import sys

import nameparser
from gooey import GooeyParser

from regolith.dates import month_to_str_int
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection

ALLOWED_TYPES = ["nsf", "doe", "other"]
ALLOWED_STATI = ["invited", "accepted", "declined", "downloaded", "inprogress", "submitted", "cancelled"]


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument("name", help="pi first name space last name in quotes", default=None)
    subpi.add_argument("type", choices=ALLOWED_TYPES, help="Report type", default=None)
    subpi.add_argument("due_date", help="Due date", **date_kwargs)
    subpi.add_argument(
        "-d",
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    subpi.add_argument("-q", "--requester", help="Name of the Program officer requesting")
    subpi.add_argument("-r", "--reviewer", help="name of the reviewer.  Defaults to sbillinge")
    subpi.add_argument("-s", "--status", choices=ALLOWED_STATI, help="Report status", default="accepted")
    subpi.add_argument("-t", "--title", help="the title of the proposal")
    return subpi


class PropRevAdderHelper(DbHelperBase):
    """Build a helper"""

    btype = "a_proprev"
    needed_colls = ["proposalReviews"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        rc.coll = "proposalReviews"
        gtx["proposalReviews"] = sorted(all_docs_from_collection(rc.client, "proposalReviews"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        name = nameparser.HumanName(rc.name)
        month = dt.datetime.today().month
        year = dt.datetime.today().year
        key = "{}{}_{}_{}".format(
            str(year)[-2:], month_to_str_int(month), name.last.casefold(), name.first.casefold().strip(".")
        )

        coll = self.gtx[rc.coll]
        pdocl = list(filter(lambda doc: doc["_id"] == key, coll))
        if len(pdocl) > 0:
            sys.exit("This entry appears to already exist in the collection")
        else:
            pdoc = {}
        pdoc.update(
            {
                "adequacy_of_resources": ["The resources available to the PI seem adequate"],
                "agency": rc.type,
                "competency_of_team": [],
                "doe_appropriateness_of_approach": [],
                "doe_reasonableness_of_budget": [],
                "doe_relevance_to_program_mission": [],
                "does_how": [],
                "does_what": "",
                "due_date": rc.due_date,
                "freewrite": [],
                "goals": [],
                "importance": [],
                "institutions": [],
                "month": month,
                "names": name.full_name,
                "nsf_broader_impacts": [],
                "nsf_create_original_transformative": [],
                "nsf_plan_good": [],
                "nsf_pot_to_advance_knowledge": [],
                "nsf_pot_to_benefit_society": [],
                "status": "accepted",
                "summary": "",
                "year": year,
            }
        )

        if rc.title:
            pdoc.update({"title": rc.title})
        else:
            pdoc.update({"title": ""})
        if rc.requester:
            pdoc.update({"requester": rc.requester})
        else:
            pdoc.update({"requester": ""})
        if rc.reviewer:
            pdoc.update({"reviewer": rc.reviewer})
        else:
            pdoc.update({"reviewer": "sbillinge"})
        if rc.status:
            if rc.status not in ALLOWED_STATI:
                raise ValueError("status should be one of {}".format(ALLOWED_STATI))
            else:
                pdoc.update({"status": rc.status})
        else:
            pdoc.update({"status": "accepted"})

        pdoc.update({"_id": key})
        rc.client.insert_one(rc.database, rc.coll, pdoc)

        print("{} proposal has been added/updated in proposal reviews".format(rc.name))

        return



"""Builder for publication lists."""

from dateutil import parser as date_parser

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.sorters import ene_date_key, position_key
from regolith.tools import all_docs_from_collection, filter_publications, make_bibtex_file

LATEX_OPTS = ["-halt-on-error", "-file-line-error"]


class PubListBuilder(LatexBuilderBase):
    btype = "publist"
    needed_colls = ["citations", "people"]

    def construct_global_ctx(self):
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if not rc.people:
            rc.people = ["all"]
        if isinstance(rc.people, str):
            rc.people = [rc.people]

        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["citations"] = sorted(
            all_docs_from_collection(rc.client, "citations"),
            key=position_key,
            reverse=True,
        )
        gtx["all_docs_from_collection"] = all_docs_from_collection

    def latex(self):
        facility = None
        filestub, qualifiers = "", ""
        if self.rc.from_date:
            from_date = date_parser.parse(self.rc.from_date).date()
            filestub = f"{filestub}_from{from_date}"
            qualifiers = f"{qualifiers} in the period from {from_date}"
            if self.rc.to_date:
                to_date = date_parser.parse(self.rc.to_date).date()
                filestub = f"{filestub}_to{to_date}"
                qualifiers = f"{qualifiers} to {to_date}"
            else:
                to_date = None
        else:
            from_date, to_date = None, None
        if self.rc.grants:
            grants = self.rc.grants
            if isinstance(grants, str):
                grants = [grants]
            if len(grants) > 2:
                text_grants = ", and ".join([",".join(grants[:-1]), grants[-1]])
                pl = "s"
            elif len(grants) == 2:
                text_grants = "and ".join([grants[0], grants[1]])
                pl = "s"
            elif len(grants) == 1:
                text_grants = grants[0]
                pl = ""
            cat_grants = ""
            for g in grants:
                cat_grants = cat_grants + "_" + g
            filestub = f"{filestub}{cat_grants}"
            qualifiers = f"{qualifiers} from Grant{pl} {text_grants}"
        if self.rc.kwargs:
            key, value = self.rc.kwargs[0].split(":", 1)
            if key == "facility":
                facility = value
                filestub = f"{filestub}_facility_{facility}"
                qualifiers = f"{qualifiers} from facility {facility}"

        for p in self.gtx["people"]:
            if p.get("_id") in self.rc.people or self.rc.people == ["all"]:
                # if self.rc.people[0] != 'all':
                #     if p.get("_id") != self.rc.people[0]:
                #         continue
                outfile = p["_id"] + filestub
                p["qualifiers"] = qualifiers
                names = frozenset(p.get("aka", []) + [p["name"]])
                citations = list(self.gtx["citations"])
                grants = self.rc.grants
                # build the bib files first without filtering for anything so they always contain all the relevant
                # publications, then
                pubs_nobold_for_bib = filter_publications(citations, names, reverse=True, bold=False, ackno=False)
                pubs_ackno_for_bib = filter_publications(citations, names, reverse=True, bold=False, ackno=True)
                pubs_for_bib = filter_publications(citations, names, reverse=True, ackno=False)
                bibfile = make_bibtex_file(pubs_for_bib, pid=p["_id"], person_dir=self.bldir)
                bibfile_nobold = make_bibtex_file(
                    pubs_nobold_for_bib, pid=f"{p['_id']}_nobold", person_dir=self.bldir
                )
                bibfile_ackno = make_bibtex_file(
                    pubs_ackno_for_bib, pid=f"{p['_id']}_ackno", person_dir=self.bldir
                )

                pubs_nobold = filter_publications(
                    citations,
                    names,
                    reverse=True,
                    bold=False,
                    ackno=False,
                    since=from_date,
                    before=to_date,
                    grants=grants,
                    facilities=facility,
                )
                pubs_ackno = filter_publications(
                    citations,
                    names,
                    reverse=True,
                    bold=False,
                    ackno=True,
                    since=from_date,
                    before=to_date,
                    grants=grants,
                    facilities=facility,
                )
                pubs = filter_publications(
                    citations,
                    names,
                    reverse=True,
                    ackno=False,
                    bold=True,
                    since=from_date,
                    before=to_date,
                    grants=grants,
                    facilities=facility,
                )

                if not p.get("email"):
                    p["email"] = ""
                emp = p.get("employment", [{"organization": ""}])
                emp.sort(key=ene_date_key, reverse=True)
                self.render(
                    "publist.tex",
                    outfile + ".tex",
                    p=p,
                    title=p.get("name", ""),
                    pubs=pubs,
                    names=names,
                    bibfile=bibfile,
                    employment=emp,
                )
                self.render(
                    "publist_nobold.tex",
                    outfile + "_nobold.tex",
                    p=p,
                    title=p.get("name", ""),
                    pubs=pubs_nobold,
                    names=names,
                    bibfile=bibfile_nobold,
                    employment=emp,
                )
                self.render(
                    "publist_ackno.tex",
                    outfile + "_ackno.tex",
                    p=p,
                    title=p.get("name", ""),
                    pubs=pubs_ackno,
                    names=names,
                    bibfile=bibfile_ackno,
                    employment=emp,
                )
                self.pdf(p["_id"])
                self.render(
                    "publist_pandoc_friendly.tex",
                    outfile + "_pandoc.tex",
                    p=p,
                    title=p.get("name", ""),
                    pubs=pubs_nobold,
                    names=names,
                    bibfile=bibfile_nobold,
                    employment=emp,
                )

    def filter_pubs_by_grant(self, pubs, grants):
        if isinstance(grants, str):
            grants = [grants]
        filtered_pubs = []
        for pub in pubs:
            for grant in grants:
                if grant in pub.get("grant", ""):
                    filtered_pubs.append(pub)
        return filtered_pubs



import json
import os
import sys
from io import StringIO

from regolith import __version__
from regolith.main import main
from regolith.runcontrol import DEFAULT_RC


def test_version():
    sys.stdout = StringIO()
    main(["--version"])
    assert sys.stdout.getvalue() == "{}\n".format(__version__)


def test_user_rc(make_db):
    repo = make_db
    DEFAULT_RC.user_config = os.path.join(repo, "user.json")
    os.chdir(repo)
    backup = sys.stdout
    sys.stdout = StringIO()
    main(["rc"])
    out1 = sys.stdout.getvalue()
    sys.stdout.close()
    sys.stdout = backup

    with open(DEFAULT_RC.user_config, "w") as f:
        json.dump({"hello": "world"}, f)

    backup = sys.stdout
    sys.stdout = StringIO()
    main(["rc"])
    out2 = sys.stdout.getvalue()
    sys.stdout.close()
    sys.stdout = backup
    assert out1 != out2
    assert "hello" in out2



import os

# from xonsh.lib import subprocess
import subprocess
import sys
from pathlib import Path

import habanero
import openpyxl
import pytest

from regolith.broker import load_db
from regolith.main import main

builder_map = [
    "cv",
    "review-prop",
    "annual-activity",
    "beamplan",
    "current-pending",
    "figure",
    "formalletter",
    #    "html",
    "internalhtml",
    "preslist",
    "publist",
    "recent-collabs",
    "grant-report",
    "resume",
    "review-man",
    # reading-lists    need tests for this
    "reimb",
]
db_srcs = ["mongo", "fs"]

xls_check = ("B17", "B20", "B36")
recent_collabs_xlsx_check = ["A51", "B51", "C51"]


def is_same(text0: str, text1: str, ignored: list):
    """Compare the content of two text. If there are different words in text0 and text1 and the word in text0
    does not contain ignored substring, return False. Else return True."""

    def should_ignore(word):
        for w in ignored:
            if w in word:
                return True
        return False

    words0, words1 = text0.split(), text1.split()
    if len(words0) != len(words1):
        return False
    for word0, word1 in zip(words0, words1):
        if not should_ignore(word0) and word0 != word1:
            return False
    return True


def prep_figure():
    # Make latex file with some jinja2 in it
    text = r"""
    \include{ {{-get_file_path(db['groups']['ergs'], 'hello')-}}}"""
    with open("figure.tex", "w") as f:
        f.write(text)
    # make file to be loaded
    os.makedirs("fig", exist_ok=True)
    with open("fig/hello.txt", "w") as f:
        f.write("hello world")
    # load the db and register the file
    db = load_db()
    print(db.get_file_path(db["groups"]["ergs"], "hello"))
    if not db.get_file_path(db["groups"]["ergs"], "hello"):
        db.add_file(db["groups"]["ergs"], "hello", "fig/hello.txt")


@pytest.mark.skipif(sys.platform == "win32", reason="does not run on windows")
@pytest.mark.parametrize("bm", builder_map)
@pytest.mark.parametrize("db_src", db_srcs)
def test_builder(bm, db_src, make_db, make_mongodb, monkeypatch):
    # FIXME: Somehow the mongo backend failed to build figure
    # FIXME: now fs is failing to build figure

    # if db_src == "mongo" and bm == "figure":
    if bm == "figure":
        return
    if db_src == "fs":
        repo = make_db
    elif db_src == "mongo":
        if make_mongodb is False:
            pytest.skip("Mongoclient failed to start")
        else:
            repo = make_mongodb
    else:
        raise ValueError("Unknown database source: {}".format(db_src))
    os.chdir(repo)
    if bm == "figure":
        prep_figure()
    if bm == "internalhtml":
        # for some reason the mocking of the crossref call doesn't work when the
        # test is run using subprocess, so skip in this case.
        # the functionality is fully tested in test_builder_python
        pytest.skip("mocking of Crossref not working with subprocess")
    if bm == "html":
        os.makedirs("templates/static", exist_ok=True)
    if bm == "reimb" or bm == "recent-collabs":
        subprocess.run(["regolith", "build", bm, "--no-pdf", "--people", "scopatz"], check=True, cwd=repo)
    elif bm == "annual-activity":
        subprocess.run(
            ["regolith", "build", bm, "--no-pdf", "--people", "sbillinge", "--from", "2017-04-01"],
            check=True,
            cwd=repo,
        )
    elif bm == "grant-report":
        main(["build", bm, "--no-pdf", "--grant", "SymPy-1.1", "--from", "2017-04-01", "--to", "2018-03-31"])
    elif bm == "publist":
        subprocess.run(["regolith", "build", bm, "--no-pdf"], check=True, cwd=repo)
        subprocess.run(
            [
                "regolith",
                "build",
                bm,
                "--no-pdf",
                "--kwargs",
                "facility:nslsii",
                "--from",
                "2016-01-01",
                "--to",
                "2018-12-01",
                "--people",
                "scopatz",
            ],
            check=True,
            cwd=repo,
        )
    else:
        subprocess.run(["regolith", "build", bm, "--no-pdf"], check=True, cwd=repo)
    repo_dir = Path(repo)
    build_dir = repo_dir / "_build" / bm
    os.chdir(build_dir)
    # find all the files in the test outputs
    test_outputs_dir = Path(__file__).parent / "outputs" / bm
    all_test_items = test_outputs_dir.rglob("*")
    test_files = [file for file in all_test_items if file.is_file()]
    # find all the files output by the function
    all_outputs = build_dir.rglob("*")
    output_files = [file for file in all_outputs if file.is_file()]
    output_filenames = [file.name for file in output_files]
    # if there is a test output file that wasn't produced by the code, fail test
    for testfile in test_files:
        if testfile.name not in output_filenames:
            print(
                f"Expected test file {testfile} not generated by the builder." f"Files genereated: {output_files}"
            )
            assert False
    # for all other test output 'expected' files check they are the same as the program output
    for file in test_files:
        actual_file = build_dir / file.name
        # if they are tex or html, disregard certain formatting things
        html_tex_bool = False
        if file.name.endswith(".html") or file.name.endswith(".tex"):
            html_tex_bool = True
        if bm == "reimb":
            actual = openpyxl.load_workbook(actual_file)["T&B"]
            actual = [str(actual[b]) for b in xls_check]
        elif bm == "recent-collabs":
            if "nsf" in actual_file.name:
                sheet = "NSF COA Template"
            else:
                sheet = "Collaborators"
            actual = openpyxl.load_workbook(actual_file)[sheet]
            actual = [str(actual[cell]) for cell in recent_collabs_xlsx_check]
        elif html_tex_bool:
            with open(actual_file, "r") as f:
                actual = [line.strip() for line in f]
            actual = [string for string in actual if ".." not in string]
            actual = [string for string in actual if "Temp" not in string]
            actual = [string for string in actual if "/tmp/" not in string]
            actual = [string for string in actual if len(string) != 0]
        else:
            with open(actual_file, "r") as f:
                actual = f.read()
        if bm == "reimb":
            expected = openpyxl.load_workbook(file)["T&B"]
            expected = [str(expected[b]) for b in xls_check]
        elif bm == "recent-collabs":
            if "nsf" in file.name:
                sheet = "NSF COA Template"
            else:
                sheet = "Collaborators"
            expected = openpyxl.load_workbook(file)[sheet]
            expected = [str(expected[cell]) for cell in recent_collabs_xlsx_check]
        elif html_tex_bool:
            with open(file, "r") as f:
                expected = [line.strip() for line in f]
            expected = [string for string in expected if ".." not in string]
            expected = [string for string in expected if "Temp" not in string]
            expected = [string for string in expected if "/tmp/" not in string]
            expected = [string for string in expected if len(string) != 0]
        else:
            with open(file, "r") as f:
                expected = f.read()

        # Skip because of a date time in
        if file != "rss.xml":
            assert expected == actual


@pytest.mark.parametrize("db_src", db_srcs)
@pytest.mark.parametrize("bm", builder_map)
def test_builder_python(bm, db_src, make_db, make_mongodb, monkeypatch):
    # FIXME: Somehow the mongo backend failed to build figure
    # FIXME: now fs is failing to build figure
    # if db_src == "mongo" and bm == "figure":
    if bm == "figure":
        return
    if db_src == "fs":
        repo = make_db
    elif db_src == "mongo":
        if make_mongodb is False:
            pytest.skip("Mongoclient failed to start")
        else:
            repo = make_mongodb
    os.chdir(repo)
    if bm == "figure":
        prep_figure()
    if bm == "internalhtml":

        def mockreturn(*args, **kwargs):
            mock_article = {
                "message": {
                    "author": [{"given": "SJL", "family": "B"}],
                    "short-container-title": ["J Club Paper"],
                    "volume": 10,
                    "title": ["title"],
                    "issued": {"date-parts": [[1971]]},
                }
            }
            return mock_article

        monkeypatch.setattr(habanero.Crossref, "works", mockreturn)
    if bm == "html":
        os.makedirs("templates/static", exist_ok=True)
    if bm == "reimb" or bm == "recent-collabs":
        main(["build", bm, "--no-pdf", "--people", "scopatz"])
    elif bm == "annual-activity":
        main(["build", bm, "--no-pdf", "--people", "sbillinge", "--from", "2017-04-01"])
    elif bm == "grant-report":
        main(["build", bm, "--no-pdf", "--grant", "SymPy-1.1", "--from", "2017-04-01", "--to", "2018-03-31"])
    elif bm == "publist":
        main(["build", bm, "--no-pdf"])
        main(
            [
                "build",
                bm,
                "--no-pdf",
                "--kwargs",
                "facility:nslsii",
                "--from",
                "2016-01-01",
                "--to",
                "2018-12-01",
                "--people",
                "scopatz",
            ]
        )
    else:
        main(["build", bm, "--no-pdf"])
    repo_dir = Path(repo)
    build_dir = repo_dir / "_build" / bm
    os.chdir(build_dir)
    # find all the files in the test outputs
    test_outputs_dir = Path(__file__).parent / "outputs" / bm
    all_test_items = test_outputs_dir.rglob("*")
    test_files = [file for file in all_test_items if file.is_file()]
    # find all the files output by the function
    all_outputs = build_dir.rglob("*")
    output_files = [file for file in all_outputs if file.is_file()]
    output_filenames = [file.name for file in output_files]
    # if there is a test output file that wasn't produced by the code, fail test
    for testfile in test_files:
        if testfile.name not in output_filenames:
            print(
                f"Expected test file {testfile} not generated by the builder." f"Files genereated: {output_files}"
            )
            assert False
    # for all other test output 'expected' files check they are the same as the program output
    for file in test_files:
        actual_file = build_dir / file.name
        # if they are tex or html, disregard certain formatting things
        html_tex_bool = False
        if file.name.endswith(".html") or file.name.endswith(".tex"):
            html_tex_bool = True
        if bm == "reimb":
            actual = openpyxl.load_workbook(actual_file)["T&B"]
            actual = [str(actual[b]) for b in xls_check]
        elif bm == "recent-collabs":
            if "nsf" in actual_file.name:
                sheet = "NSF COA Template"
            else:
                sheet = "Collaborators"
            actual = openpyxl.load_workbook(actual_file)[sheet]
            actual = [str(actual[cell]) for cell in recent_collabs_xlsx_check]
        elif html_tex_bool:
            with open(actual_file, "r") as f:
                actual = [line.strip() for line in f]
            actual = [string for string in actual if ".." not in string]
            actual = [string for string in actual if "Temp" not in string]
            actual = [string for string in actual if "/tmp/" not in string]
            actual = [string for string in actual if len(string) != 0]
        else:
            with open(actual_file, "r") as f:
                actual = f.read()

        if bm == "reimb":
            expected = openpyxl.load_workbook(file)["T&B"]
            expected = [str(expected[b]) for b in xls_check]
        elif bm == "recent-collabs":
            if "nsf" in file.name:
                sheet = "NSF COA Template"
            else:
                sheet = "Collaborators"
            expected = openpyxl.load_workbook(file)[sheet]
            expected = [str(expected[cell]) for cell in recent_collabs_xlsx_check]
        elif html_tex_bool:
            with open(file, "r") as f:
                expected = [line.strip() for line in f]
            expected = [string for string in expected if ".." not in string]
            expected = [string for string in expected if "Temp" not in string]
            expected = [string for string in expected if "/tmp/" not in string]
            expected = [string for string in expected if len(string) != 0]
        else:
            with open(file, "r") as f:
                expected = f.read()

        # Skip because of a date time in
        if file != "rss.xml":
            assert expected == actual



"""Helps deploy what we have built."""

import os
import time
from distutils.dir_util import copy_tree
from warnings import warn

from xonsh.lib import subprocess

try:
    import hglib
except ImportError:
    hglib = None


def ensure_deploy_dir(rc):
    """Ensure deployment dir is on rc and physically exists."""
    if not hasattr(rc, "deploydir") or rc.deploydir is None:
        rc.deploydir = os.path.join(rc.builddir, "deploy")
    if not os.path.isdir(rc.deploydir):
        os.makedirs(rc.deploydir, exist_ok=True)


def deploy_git(rc, name, url, src="html", dst=None):
    """Loads a git database"""
    targetdir = os.path.join(rc.deploydir, name)
    # get or update the database
    if os.path.isdir(targetdir):
        cmd = ["git", "pull"]
        cwd = targetdir
    else:
        cmd = ["git", "clone", url, targetdir]
        cwd = None
    subprocess.check_call(cmd, cwd=cwd)
    # copy the files over
    srcdir = os.path.join(rc.builddir, src)
    dstdir = os.path.join(targetdir, dst) if dst else targetdir
    copy_tree(srcdir, dstdir, verbose=1)
    # commit everything
    cmd = ["git", "add", "."]
    subprocess.check_call(cmd, cwd=targetdir)
    # commit
    cmd = [
        "git",
        "commit",
        "-m",
        "regolith auto-deploy at {0}".format(time.time()),
    ]
    try:
        subprocess.check_call(cmd, cwd=targetdir)
    except subprocess.CalledProcessError:
        warn("Could not git commit to " + targetdir, RuntimeWarning)
        # don't return, just in case...
    # deploy!
    cmd = ["git", "push"]
    try:
        subprocess.check_call(cmd, cwd=targetdir)
    except subprocess.CalledProcessError:
        warn("Could not git push from " + targetdir, RuntimeWarning)
        return


def deploy_hg(rc, name, url, src="html", dst=None):
    """Loads an hg database"""
    if hglib is None:
        raise ImportError("hglib")
    targetdir = os.path.join(rc.deploydir, name)
    # get or update the database
    if os.path.isdir(targetdir):
        client = hglib.open(targetdir)
        client.pull(update=True)
    else:
        # Strip off hg+
        hglib.clone(url[3:], targetdir)
        client = hglib.open(targetdir)
    # copy the files over
    srcdir = os.path.join(rc.builddir, src)
    dstdir = os.path.join(targetdir, dst) if dst else targetdir
    copy_tree(srcdir, dstdir, verbose=1)
    # commit everything
    client.commit(
        message="regolith auto-deploy at {0}".format(time.time()),
        addremove=True,
    )
    client.push()


def deploy(rc, name, url, src="html", dst=None):
    """Deploys a target"""
    ensure_deploy_dir(rc)
    if url.startswith("git") or url.endswith(".git"):
        deploy_git(rc, name, url, src=src, dst=dst)
    elif url.startswith("hg+"):
        deploy_hg(rc, name, url, src=src, dst=dst)
    else:
        raise ValueError("Do not know how to deploy to this kind of URL: " + url)



"""Builder for Proposal Reivews."""

from nameparser import HumanName

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.fsclient import _id_key
from regolith.tools import all_docs_from_collection, dereference_institution


class PropRevBuilder(LatexBuilderBase):
    """Build a proposal review from database entries"""

    btype = "review-prop"
    needed_colls = ["institutions", "proposalReviews"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["proposalReviews"] = sorted(all_docs_from_collection(rc.client, "proposalReviews"), key=_id_key)
        gtx["institutions"] = sorted(all_docs_from_collection(rc.client, "institutions"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        """Render latex template"""
        for rev in self.gtx["proposalReviews"]:
            outname = "{}_{}".format(_id_key(rev), rev["reviewer"])
            multiauth = False
            if isinstance(rev["names"], str):
                rev["names"] = [rev["names"]]
            if len(rev["names"]) > 1:
                multiauth = True
            firstauthor = HumanName(rev["names"][0])
            firstauthorlastname = firstauthor.last

            if isinstance(rev.get("institutions", ""), str):
                rev["institutions"] = [rev.get("institutions", "")]
            if len(rev["institutions"]) == 0:
                rev["institutions"] = [""]
            institution_docs = []
            for inst in rev["institutions"]:
                instdoc = {"institution": inst}
                dereference_institution(instdoc, self.gtx["institutions"])
                if not instdoc:
                    instdoc = {"institution": inst}
                institution_docs.append(instdoc)
            if isinstance(rev["freewrite"], str):
                rev["freewrite"] = [rev["freewrite"]]

            self.render(
                "propreport.txt",
                "{}.txt".format(outname),
                trim_blocks=True,
                agency=rev["agency"],
                appropriateness=rev["doe_appropriateness_of_approach"],
                title=rev["title"],
                institution=institution_docs[0].get("institution"),
                multiauthor=multiauth,
                firstAuthorLastName=firstauthorlastname,
                competency=rev["competency_of_team"],
                adequacy=rev["adequacy_of_resources"],
                does_what=rev["does_what"],
                relevance=rev["doe_relevance_to_program_mission"],
                budget=rev["doe_reasonableness_of_budget"],
                does_how=rev["does_how"],
                goals=rev["goals"],
                importance=rev["importance"],
                summary=f"{rev['summary']}\n",
                freewrite=rev["freewrite"],
                broader_impacts=rev["nsf_broader_impacts"],
                creativity_originality=rev["nsf_create_original_transformative"],
                benefit_to_society=rev["nsf_pot_to_benefit_society"],
                plan_good=rev["nsf_plan_good"],
                advance_knowledge=rev["nsf_pot_to_advance_knowledge"],
            )



"""Helper for adding a presentation to the presentation collection."""

import time
from warnings import warn

import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.a_expensehelper import expense_constructor
from regolith.helpers.basehelper import DbHelperBase
from regolith.schemas import alloweds
from regolith.tools import (
    add_to_google_calendar,
    all_docs_from_collection,
    create_repo,
    get_pi_id,
    google_cal_auth_flow,
)

TARGET_COLL = "presentations"
EXPENSES_COLL = "expenses"

PRESENTATION_TYPES = alloweds.get("PRESENTATION_TYPES")
PRESENTATION_STATI = alloweds.get("PRESENTATION_STATI")


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument(
        "name",
        help="name of the event of the presentation. Meeting name if meeting, " "department if seminar",
    )
    subpi.add_argument(
        "place", help="the place of the presentation, Location if conference, " "institution for seminars"
    )
    subpi.add_argument("begin_date", help="Input begin date for this presentation ", **date_kwargs)
    subpi.add_argument("end_date", help="Input end date for this presentation", **date_kwargs)
    subpi.add_argument("-t", "--title", help="the title of the presentation, default is tbd", default="tbd")
    subpi.add_argument("-a", "--abstract", help="abstract of the presentation, defaults to tbd", default="tbd")
    subpi.add_argument(
        "-p",
        "--person",
        help="the person presenting the presentation, used for presentation name,"
        " defaults to name in user.config",
    )
    subpi.add_argument(
        "--authors",
        nargs="+",
        help="specify the authors of this presentation, " "defaults to person submitting the presentation",
    )
    subpi.add_argument(
        "-g",
        "--grants",
        nargs="+",
        help="grant, or grants (separated by spaces), that support this presentation. Defaults to tbd",
        default="tbd",
    )
    subpi.add_argument(
        "-u",
        "--presentation-url",
        help="the url to the presentation, whether it is on Google Drive, GitHub or wherever",
    )
    subpi.add_argument(
        "-n",
        "--notes",
        nargs="+",
        help="note or notes to be inserted as a list into the notes field, "
        "separate notes with spaces.  Place inside quotes if the note "
        "itself contains spaces.",
        default=[],
    )
    subpi.add_argument(
        "-s",
        "--status",
        choices=PRESENTATION_STATI,
        help=f"status, from {PRESENTATION_STATI}, default is accepted",
        default="accepted",
    )
    subpi.add_argument(
        "-y",
        "--type",
        choices=PRESENTATION_TYPES,
        help=f"types, from {PRESENTATION_TYPES}. Default is invited",
        default="invited",
    )
    subpi.add_argument("-w", "--webinar", help="Is the presentation a webinar?", action="store_true")
    subpi.add_argument(
        "--no-expense",
        help="Do not add a template expense item to the "
        "expenses collection.  Default is to add "
        "an expense if the presentation is not a "
        "webinar.",
        action="store_true",
    )
    subpi.add_argument(
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    subpi.add_argument(
        "--expense-db",
        help="The database where the expense collection will be updated "
        "with the presentation-related expense data. "
        "Without a specification, by default the helper attempts to "
        "use the first database in the regolithrc.json file *but "
        "will only do so if that database is non-public*.",
    )
    subpi.add_argument(
        "--force",
        help="Force adding presentation expense data to the first DB listed in "
        "'databases' in the regolithrc.json file, *even if that database is "
        "public*. DANGER: This could reveal sensitive information to the public.",
        action="store_true",
    )
    subpi.add_argument(
        "--id",
        help="Override the default id created from the date, " "speaker and place by specifying an id here",
    )
    subpi.add_argument("--no-cal", help="Do not add the presentation to google calendar", action="store_true")
    subpi.add_argument(
        "--no-repo", help="Do not create a GitHub/Lab repo for the presentation", action="store_true"
    )
    return subpi


class PresentationAdderHelper(DbHelperBase):
    """Helper for adding presentations" """

    # btype must be the same as helper target in helper.py
    btype = "a_presentation"
    needed_colls = [f"{TARGET_COLL}", "groups", "people", "expenses"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        if not rc.database:
            rc.database = rc.databases[0]["name"]

        if rc.no_expense:
            if rc.expense_db:
                raise RuntimeError(
                    "ERROR: You specified an expense database with the --expense-db option, but also "
                    "passed --no-expense. Do you want to create an expense? Please reformulate your "
                    "helper command and rerun. "
                )
        else:
            rc.expense_db = rc.expense_db if rc.expense_db else None

            if not rc.expense_db and not rc.databases[0].get("public"):
                warn(
                    "WARNING: No expense database was provided to input the expense data "
                    "associated with this presentation. Defaulted to using the first DB "
                    "listed in your regolithrc.json, as this DB is non-public."
                )
                rc.expense_db = rc.databases[0]["name"]

            rc.expense_db = (
                rc.databases[0]["name"]
                if not rc.expense_db and rc.databases[0].get("public") and rc.force
                else rc.expense_db
            )

            if not rc.expense_db and rc.databases[0].get("public") and not rc.force:
                raise RuntimeError(
                    "ERROR: You failed to specify an expense database for the expense data "
                    "associated with this presentation. The helper defaults to entering "
                    "expenses into first database listed in your regolithrc.json file, "
                    "but it set to PUBLIC and would reveal potentially sensitive information. "
                    "Rerun by specifying the target database with --expense-db EXPENSE_DB, or "
                    "(at your own risk) pass the --force flag."
                )

            if rc.expense_db not in [database.get("name") for database in rc.databases]:
                raise RuntimeError(
                    f"ERROR: The expense database specified, {rc.expense_db}, is not listed "
                    "in your regolithrc.json file."
                )

        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc

        if not rc.no_cal:
            event = {
                "summary": rc.name,
                "location": rc.place,
                "start": {"date": rc.begin_date},
                "end": {"date": rc.end_date},
            }
            cal_update_bool = add_to_google_calendar(event)
            if not cal_update_bool:
                google_cal_auth_flow()
                wait_bool, jump = 0, 0
                while not wait_bool or jump == 60:
                    time.sleep(1)
                    wait_bool = add_to_google_calendar(event)
                    jump += 1

        # dates
        # TODO : add date format check?
        begin_date = date_parser.parse(rc.begin_date).date()
        end_date = date_parser.parse(rc.end_date).date()

        # User specifies person in CL or in config.json
        if not rc.person:
            try:
                rc.person = rc.default_user_id
            except AttributeError:
                raise RuntimeError(
                    "WARNING: no person been set. please rerun specifying authors,"
                    "or add your id, e.g., sbillinge, to the config.json file in ~/.config/regolith"
                )
        split_person = rc.person.strip().split()
        if len(split_person) >= 2:
            name_key = split_person[0][0].casefold() + split_person[-1][0].casefold()
        else:
            name_key = split_person[0][:2].casefold()

        if not rc.id:
            key = (
                f"{str(begin_date.year)[2:]}{str(begin_date.strftime('%m'))}{name_key}_"
                f"{''.join(rc.place.casefold().split()).strip()}"
            )
        else:
            key = rc.id

        coll = self.gtx[rc.coll]
        pdocl = list(filter(lambda doc: doc["_id"] == key, coll))
        if len(pdocl) > 0:
            raise RuntimeError("This entry appears to already exist in the collection")
        else:
            pdoc = {}

        if not rc.authors:
            authors = [rc.person]
        else:
            authors = rc.authors

        pdoc.update(
            {
                "_id": key,
                "abstract": rc.abstract,
                "authors": authors,
                "begin_date": begin_date,
                "end_date": end_date,
                "notes": rc.notes,
            }
        )
        if rc.presentation_url:
            pdoc.update({"presentation_url": rc.presentation_url})
        if rc.webinar:
            rc.no_expense = True
            pdoc.update({"webinar": True})
        if rc.type in ["seminar", "colloquium"]:
            pdoc.update({"institution": rc.place, "department": rc.name})
        else:
            pdoc.update({"location": rc.place, "meeting_name": rc.name})
        pdoc.update(
            {
                "project": ["all"],
                "status": rc.status,
                "title": rc.title,
                "type": rc.type,
            }
        )

        rc.client.insert_one(rc.database, rc.coll, pdoc)
        print(f"{key} has been added in {TARGET_COLL}")

        if not rc.no_expense:
            rc.business = False
            rc.payee = rc.default_user_id
            rc.purpose = f"give {rc.type} presentation at {rc.name}, {rc.place}"
            rc.where = "tbd"
            rc.status = "unsubmitted"
            edoc = expense_constructor(key, begin_date, end_date, rc)
            rc.client.insert_one(rc.expense_db, EXPENSES_COLL, edoc)
            print(f"{key} has been added in {EXPENSES_COLL} in database {rc.expense_db}")

        if not rc.no_repo:
            if not hasattr(rc, "repos"):
                rc.repos = []
            if not hasattr(rc, "tokens"):
                rc.tokens = []
            for repo in rc.repos:
                if repo.get("_id") == "talk_repo":
                    repo["params"].update({"name": key})
                    if not repo["params"].get("initialize_with_readme"):
                        repo["params"]["initialize_with_readme"] = True
                    repo["params"].update({"name": key.replace("/", "").replace(",", "")})
            msg = create_repo("talk_repo", "gitlab_private_token", rc)
            print(msg)
        return



"""Flask app for grading regolith."""

import json
import traceback

from flask import Flask, abort, render_template, request

app = Flask("regolith")


@app.route("/", methods=["GET", "POST"])
def root():
    rc = app.rc
    status = None
    if request.method == "POST":
        form = request.form
        if "shutdown" in form:
            return shutdown()
        if "rowdata" in form:
            grade = form_to_grade_row(form)
        else:
            grade = form_to_grade_assignment(form)
        insert_grade(grade, form, rc)
        status = "submitted {0} ✓".format(grade["_id"])
        return status
    return render_template(
        "grader.html",
        json=json,
        rc=rc,
        str=str,
        status=status,
        range=range,
        len=len,
        sorted=sorted,
        enumerate=enumerate,
        by_id=lambda x: x["_id"],
    )


def shutdown_server():
    func = request.environ.get("werkzeug.server.shutdown")
    if func is None:
        raise RuntimeError("Not running with the Werkzeug Server")
    func()


@app.route("/shutdown", methods=["GET", "POST"])
def shutdown():
    shutdown_server()
    return "Regolith server shutting down...\n"


def form_to_grade_assignment(form):
    """Creates a grade dict from an assignment form."""
    grade_id = "{student}-{assignment}-{course}".format(**form)
    grade = {
        "_id": grade_id,
        "student": form["student"],
        "assignment": form["assignment"],
        "course": form["course"],
    }
    if form["filename"]:
        grade["filename"] = form["filename"]
    scores = {int(k[5:]): float(v) for k, v in form.items() if k.startswith("score")}
    scores = sorted(scores.items())
    grade["scores"] = [v for _, v in scores]
    return grade


def form_to_grade_row(form):
    """Creates a grade dict from a row form."""
    course = form["course"]
    student = form["student"]
    assignment, _, _ = form["assignment"].partition("[")
    rowdata = json.loads(form["rowdata"])
    grade = {"student": student, "assignment": assignment, "course": course}
    grade["_id"] = "{student}-{assignment}-{course}".format(**grade)
    scores = {}
    for k, v in rowdata.items():
        if not k.startswith(assignment):
            continue
        _, _, n = k.partition("[")
        n, _, _ = n.partition("]")
        scores[int(n)] = v
    scores = sorted(scores.items())
    grade["scores"] = [v for _, v in scores]
    return grade


def insert_grade(grade, form, rc):
    """Inserts a grade into the database."""
    dbname = form["dbname"]
    collname = "grades"
    try:
        rc.client[dbname][collname]
    except (KeyError, AttributeError):
        abort(404)
    try:
        rc.client.insert_one(dbname, collname, grade)
    except Exception:
        traceback.print_exc()
        raise



import json
import os
import shutil
import sys
import tempfile
from copy import deepcopy

from xonsh.lib import subprocess

from regolith.broker import load_db
from regolith.fsclient import dump_yaml
from regolith.main import main
from regolith.schemas import EXEMPLARS

builder_map = [
    "cv",
    "html",
    "resume",
    "publist",
    "current-pending",
    "preslist",
    "figure",
    "reimb",
]


def prep_figure():
    # Make latex file with some jinja2 in it
    text = r"""
    \include{ {{-get_file_path(db['groups']['ergs'], 'hello')-}}}"""
    with open("figure.tex", "w") as f:
        f.write(text)
    # make file to be loaded
    os.makedirs("fig", exist_ok=True)
    with open("fig/hello.txt", "w") as f:
        f.write("hello world")
    # load the db and register the file
    db = load_db()
    print(db.get_file_path(db["groups"]["ergs"], "hello"))
    if not db.get_file_path(db["groups"]["ergs"], "hello"):
        db.add_file(db["groups"]["ergs"], "hello", "fig/hello.txt")


def rmtree(dirname):
    """Remove a directory, even if it has read-only files (Windows).
    Git creates read-only files that must be removed on teardown. See
    https://stackoverflow.com/questions/2656322  for more info.

    Parameters
    ----------
    dirname : str
        Directory to be removed
    """
    try:
        shutil.rmtree(dirname)
    except PermissionError:
        if sys.platform == "win32":
            subprocess.check_call(["del", "/F/S/Q", dirname], shell=True)
        else:
            raise


def make_db():
    """A test fixutre that creates and destroys a git repo in a temporary
    directory.
    This will yield the path to the repo.
    """
    cwd = os.getcwd()
    name = "regolith_fake"
    repo = os.path.join(tempfile.gettempdir(), name)
    if os.path.exists(repo):
        rmtree(repo)
    subprocess.run(["git", "init", repo])
    os.chdir(repo)
    with open("README", "w") as f:
        f.write("testing " + name)
    with open("regolithrc.json", "w") as f:
        json.dump(
            {
                "groupname": "ERGS",
                "databases": [
                    {
                        "name": "test",
                        "url": repo,
                        "public": True,
                        "path": "db",
                        "local": True,
                    }
                ],
                "stores": [
                    {
                        "name": "store",
                        "url": repo,
                        "path": repo,
                        "public": True,
                    }
                ],
                "force": False,
            },
            f,
        )
    os.mkdir("db")
    # Write collection docs
    for coll, example in deepcopy(EXEMPLARS).items():
        if isinstance(example, list):
            d = {dd["_id"]: dd for dd in example}
        else:
            d = {example["_id"]: example}
        dump_yaml("db/{}.yaml".format(coll), d)
    subprocess.run(["git", "add", "."])
    subprocess.run(["git", "commit", "-am", "Initial readme"])
    yield repo
    os.chdir(cwd)
    rmtree(repo)


expected_base = os.path.join(os.path.abspath(os.path.dirname(__file__)), "outputs")


def bootstrap_builders():
    g = make_db()
    repo = next(g)
    os.chdir(repo)
    for bm in builder_map:
        if bm == "html":
            os.makedirs("templates/static")
        if bm == "figure":
            prep_figure()
        main(["build", bm, "--no-pdf"])
        os.chdir(os.path.join(repo, "_build", bm))
        os.makedirs(expected_base, exist_ok=True)
        for root, dirs, files in os.walk("."):
            for file in files:
                # Use this for bootstrapping the tests,
                # confirm by hand that files look correct
                if root == ".":
                    os.makedirs(os.path.join(expected_base, bm), exist_ok=True)
                    shutil.copyfile(
                        os.path.join(file),
                        os.path.join(expected_base, bm, file),
                    )
                else:
                    os.makedirs(os.path.join(expected_base, bm, root), exist_ok=True)
                    shutil.copyfile(
                        os.path.join(root, file),
                        os.path.join(expected_base, bm, root, file),
                    )
        os.chdir(repo)
    next(g)


if __name__ == "__main__":
    bootstrap_builders()



"""Tools for document storgage."""

import os
import shutil
from contextlib import contextmanager
from warnings import warn

from xonsh.lib import subprocess

try:
    import hglib
except ImportError:
    hglib = None


def find_store(rc):
    if getattr(rc, "storename", None) is None and len(rc.stores) != 0:
        return rc.stores[0]
    for store in rc.stores:
        if store["name"] == rc.storename:
            return store
    else:
        msg = "Could not find the store {0!r}".format(rc.storename)
        raise RuntimeError(msg)


def storage_path(store, rc):
    """Computes the storage directory."""
    name, url = store["name"], store["url"]
    for db in rc.databases:
        if db["name"] == name and db["url"] == url:
            path = os.path.join(rc.builddir, "_dbs", name, store["path"])
            break
    else:
        path = os.path.join(rc.builddir, "_stores", name, store["path"])
    os.makedirs(path, exist_ok=True)
    return path


def sync_git(store, path):
    """Syncs the local documents via git."""
    storedir, _ = os.path.split(path)
    # get or update the storage
    if os.path.isdir(storedir):
        cmd = ["git", "pull"]
        cwd = storedir
    else:
        cmd = ["git", "clone", store["url"], storedir]
        cwd = None
    subprocess.check_call(cmd, cwd=cwd)


def sync_hg(store, path):
    """Syncs the local documents via hg."""
    storedir, _ = os.path.split(path)
    # get or update the storage
    if os.path.isdir(os.path.join(storedir, ".hg")):
        client = hglib.open(storedir)
        client.pull(update=True, force=True)
    else:
        # Strip off three characters for hg+
        client = hglib.clone(store["url"][3:], storedir)


def sync(store, path):
    """Syncs the local documents."""
    url = store["url"]
    if url.startswith("git") or url.endswith(".git"):
        sync_git(store, path)
    elif url.startswith("hg+"):
        sync_hg(store, path)
    elif not os.path.exists(os.path.expanduser(url)):
        raise ValueError("Do not know how to sync this kind of storage.")


def copydocs(store, path, rc):
    """Copies files to the staging area."""
    for doc in rc.documents:
        dst = os.path.join(path, os.path.split(doc)[1])
        if not rc.force and os.path.isfile(dst):
            raise RuntimeError(dst + " already exists!")
        shutil.copy2(doc, dst)


def push_git(store, path):
    """Pushes the local documents via git."""
    storedir, _ = os.path.split(path)
    cmd = ["git", "add", "."]
    subprocess.check_call(cmd, cwd=storedir)
    cmd = ["git", "commit", "-m", "regolith auto-store commit"]
    try:
        subprocess.check_call(cmd, cwd=storedir)
    except subprocess.CalledProcessError:
        warn("Could not git commit to " + storedir, RuntimeWarning)
        return
    cmd = ["git", "push"]
    try:
        subprocess.check_call(cmd, cwd=storedir)
    except subprocess.CalledProcessError:
        warn("Could not git push from " + storedir, RuntimeWarning)
        return


def push_hg(store, path):
    """Pushes the local documents via git."""
    storedir, _ = os.path.split(path)
    client = hglib.open(storedir)
    if len(client.status(modified=True, unknown=True, added=True)) == 0:
        return
    client.commit(message="regolith auto-commit", addremove=True)
    client.push()


def push(store, path):
    """Pushes the local documents."""
    url = store["url"]
    if url.startswith("git") or url.endswith(".git"):
        push_git(store, path)
    elif url.startswith("hg+"):
        push_hg(store, path)
    elif not os.path.exists(os.path.expanduser(url)):
        raise ValueError("Do not know how to push to this kind of storage.")


class StorageClient(object):
    """Interface to the storage system"""

    def __init__(self, rc, store, path):
        self.rc = rc
        self.store = store
        self.path = path

    def copydoc(self, doc):
        """Copies file to the staging area."""
        dst = os.path.join(self.path, os.path.split(doc)[1])
        if not self.rc.force and os.path.isfile(dst):
            raise RuntimeError(dst + " already exists!")
        shutil.copy2(doc, dst)
        return dst

    def retrieve(self, file_name):
        """Get file from the store

        Parameters
        ----------
        file_name : name of the file

        Returns
        -------
        path : str or None
            The path, if the file is not in the store None
        """
        ret = os.path.join(self.path, file_name)
        temp = (self.path + file_name).find(self.path, 1, -1)
        if os.path.exists(ret):
            return os.path.join(self.path, file_name)
        elif temp != -1:
            if os.name == "posix":
                return os.getcwd() + "/" + file_name
            else:
                return os.getcwd() + "\\" + file_name
        else:
            return None


@contextmanager
def store_client(rc):
    """Context manager for file storage

    Parameters
    ----------
    rc : RunControl

    Yields
    -------
    client : StorageClient
        The StorageClient instance
    """
    store = find_store(rc)
    path = storage_path(store, rc)
    sync(store, path)
    yield StorageClient(rc, store, path)
    push(store, path)


def main(rc):
    """Copies files into the local storage location and uploads them."""
    with store_client(rc) as sclient:
        for doc in rc.documents:
            sclient.copydoc(doc)



"""Builder for Lists of Presentations.

This builder will build a presentation list for each group-member in each group
listed in groups.yml.  Group members are indicated in the employment and
education sections of the individual in people.yml.

There are a number of filtering options, i.e., for presentation-type (invited,
colloquium, seminar, poster, contributed_oral) and for whether the invitation
was accepted or declined.  As of now, these filters can only be updated by
editing this file but may appear as command-line options later.  It will also
get institution and department information from institutions.yml if they are
there.

The author list is built from information in people.yml where possible and
contacts.yml as a fallback.  The author list is built by first performing a
fuzzy search for the person in people.yml, followed by a fuzzy search through
contacts.yml (if the person was not found in people.yml), but if the person is
absent both database files, it will still build but using the string name given
in the presentations.yml. The presentations are output in a ./_build directory."""

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.fsclient import _id_key
from regolith.sorters import position_key
from regolith.stylers import month_fullnames, sentencecase
from regolith.tools import all_docs_from_collection, filter_presentations, group_member_ids


class PresListBuilder(LatexBuilderBase):
    """Build list of talks and posters (presentations) from database entries"""

    btype = "preslist"
    needed_colls = ["groups", "institutions", "people", "grants", "presentations", "contacts"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["contacts"] = sorted(
            all_docs_from_collection(rc.client, "contacts"),
            key=position_key,
            reverse=True,
        )
        gtx["grants"] = sorted(all_docs_from_collection(rc.client, "grants"), key=_id_key)
        gtx["groups"] = sorted(all_docs_from_collection(rc.client, "groups"), key=_id_key)
        gtx["presentations"] = sorted(all_docs_from_collection(rc.client, "presentations"), key=_id_key)
        gtx["institutions"] = sorted(all_docs_from_collection(rc.client, "institutions"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        """Render latex template"""
        everybody = self.gtx["people"] + self.gtx["contacts"]
        for group in self.gtx["groups"]:
            grp = group["_id"]
            grpmember_ids = group_member_ids(self.gtx["people"], grp)
            for member in grpmember_ids:
                if self.rc.people:
                    if member not in self.rc.people:
                        continue
                presclean = filter_presentations(
                    everybody, self.gtx["presentations"], self.gtx["institutions"], member, statuses=["accepted"]
                )

                if len(presclean) > 0:
                    presclean = sorted(
                        presclean,
                        key=lambda k: k.get("date", None),
                        reverse=True,
                    )
                    outfile = "presentations-" + grp + "-" + member
                    pi = [person for person in self.gtx["people"] if person["_id"] is member][0]
                    self.render(
                        "preslist.tex",
                        outfile + ".tex",
                        pi=pi,
                        presentations=presclean,
                        sentencecase=sentencecase,
                        monthstyle=month_fullnames,
                    )
                    self.env.trim_blocks = True
                    self.env.lstrip_blocks = True
                    self.render(
                        "preslist.txt",
                        outfile + ".txt",
                        pi=pi,
                        presentations=presclean,
                        sentencecase=sentencecase,
                        monthstyle=month_fullnames,
                    )
                    self.pdf(outfile)



import datetime
from datetime import date

import pytest

from regolith.dates import (
    date_to_float,
    day_to_str_int,
    find_gaps_overlaps,
    get_dates,
    get_due_date,
    has_finished,
    has_started,
    is_after,
    is_before,
    is_between,
    is_current,
    last_day,
    month_to_int,
    month_to_str_int,
)

TEST_DATE = date(2019, 6, 15)
TEST_START_DATE = date(2019, 1, 1)
TEST_END_DATE = date(2019, 2, 5)


@pytest.mark.parametrize(
    "input,expected",
    [
        (1, "01"),
        (10, "10"),
        ("Oct", "10"),
        ("Jan", "01"),
    ],
)
def test_month_to_str(input, expected):
    assert month_to_str_int(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        ("2020-05-01", datetime.date(2020, 5, 1)),
        (datetime.date(2020, 5, 1), datetime.date(2020, 5, 1)),
        (2020, True),
    ],
)
def test_get_due_date(input, expected):
    with pytest.raises(Exception):
        assert get_due_date(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [("Jan", 1), (1, 1), ("February", 2)],
)
def test_month_to_int(input, expected):
    assert month_to_int(input) == expected


@pytest.mark.parametrize(
    "input,expected",
    [([2019, 1, 15], 2019.0115), ([2019, "May", 0], 2019.05), ([2019, "February", 2], 2019.0202)],
)
def test_date_to_float(input, expected):
    assert date_to_float(input[0], input[1], d=input[2]) == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (1, "01"),
        (10, "10"),
    ],
)
def test_day_to_str(input, expected):
    assert day_to_str_int(input) == expected


@pytest.mark.parametrize(
    "input,flag,expected",
    [
        ([(date(2020, 1, 1), date(2020, 1, 31)), (date(2020, 2, 1), date(2020, 2, 5))], False, True),
        ([(date(2020, 2, 1), date(2020, 2, 5)), (date(2020, 1, 1), date(2020, 1, 31))], False, True),
        ([(date(2020, 1, 1), date(2020, 1, 31)), (date(2020, 2, 2), date(2020, 2, 5))], False, False),
        ([(date(2020, 1, 1), date(2020, 1, 31)), (date(2020, 1, 31), date(2020, 2, 5))], False, False),
        ([(date(2020, 1, 1), date(2020, 1, 31)), (date(2020, 1, 31), date(2020, 2, 5))], True, True),
        (
            [
                (date(2020, 1, 1), date(2020, 1, 31)),
                (date(2020, 2, 1), date(2020, 2, 5)),
                (date(2020, 2, 6), date(2020, 2, 7)),
            ],
            False,
            True,
        ),
        (
            [
                (date(2020, 1, 1), date(2020, 1, 31)),
                (date(2020, 2, 1), date(2020, 2, 5)),
                (date(2020, 2, 7), date(2020, 2, 7)),
            ],
            False,
            False,
        ),
    ],
)
def test_find_gaps_overlaps(input, flag, expected):
    actual = find_gaps_overlaps(input, overlaps_ok=flag)
    assert actual == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (
            ({"year": 2020}, None),
            {"begin_date": datetime.date(2020, 1, 1), "end_date": datetime.date(2020, 12, 31)},
        ),
        (
            ({"year": 2020, "month": 9}, None),
            {"begin_date": datetime.date(2020, 9, 1), "end_date": datetime.date(2020, 9, 30)},
        ),
        (({}, None), {}),
        (({"year": "tbd", "month": 9}, None), {}),
        (
            ({"year": 2019, "month": "tbd"}, None),
            {"begin_date": datetime.date(2019, 1, 1), "end_date": datetime.date(2019, 12, 31)},
        ),
        (
            ({"year": 2020, "month": 9}, None),
            {"begin_date": datetime.date(2020, 9, 1), "end_date": datetime.date(2020, 9, 30)},
        ),
        (
            ({"year": 2020, "month": "Sep", "day": 15}, None),
            {
                "begin_date": datetime.date(2020, 9, 15),
                "end_date": datetime.date(2020, 9, 15),
                "date": datetime.date(2020, 9, 15),
            },
        ),
        (({"begin_year": 2020}, None), {"begin_date": datetime.date(2020, 1, 1)}),
        (({"begin_year": 2020, "begin_month": 4}, None), {"begin_date": datetime.date(2020, 4, 1)}),
        (
            ({"begin_year": 2020, "begin_month": 4, "begin_day": 5}, None),
            {"begin_date": datetime.date(2020, 4, 5)},
        ),
        (
            ({"begin_year": "2020", "begin_month": "4", "begin_day": "5"}, None),
            {"begin_date": datetime.date(2020, 4, 5)},
        ),
        (
            ({"begin_year": 2019, "end_year": 2020}, None),
            {"begin_date": datetime.date(2019, 1, 1), "end_date": datetime.date(2020, 12, 31)},
        ),
        (
            ({"begin_year": 2019, "end_year": 2020, "end_month": "Feb"}, None),
            {"begin_date": datetime.date(2019, 1, 1), "end_date": datetime.date(2020, 2, 29)},
        ),
        (
            ({"begin_year": 2019, "end_year": 2020, "end_month": "Feb", "end_day": 10}, None),
            {"begin_date": datetime.date(2019, 1, 1), "end_date": datetime.date(2020, 2, 10)},
        ),
        (
            (
                {
                    "begin_date": "2020-05-09",
                    "begin_year": 2019,
                    "end_year": 2020,
                    "end_month": "Feb",
                    "end_day": 10,
                },
                None,
            ),
            {"begin_date": datetime.date(2020, 5, 9), "end_date": datetime.date(2020, 2, 10)},
        ),
        (
            (
                {"end_date": "2020-5-20", "begin_year": 2019, "end_year": 2020, "end_month": "Feb", "end_day": 10},
                None,
            ),
            {"begin_date": datetime.date(2019, 1, 1), "end_date": datetime.date(2020, 5, 20)},
        ),
        (
            ({"date": "2020-5-20", "begin_year": 2019, "end_year": 2020, "end_month": "Feb", "end_day": 10}, None),
            {
                "begin_date": datetime.date(2019, 1, 1),
                "end_date": datetime.date(2020, 2, 10),
                "date": datetime.date(2020, 5, 20),
            },
        ),
        (
            (
                {
                    "date": datetime.date(2020, 5, 20),
                    "begin_year": 2019,
                    "end_year": 2020,
                    "end_month": "Feb",
                    "end_day": 10,
                },
                None,
            ),
            {
                "begin_date": datetime.date(2019, 1, 1),
                "end_date": datetime.date(2020, 2, 10),
                "date": datetime.date(2020, 5, 20),
            },
        ),
        (
            (
                {
                    "date": datetime.date(2020, 5, 20),
                    "begin_date": datetime.date(2015, 6, 8),
                    "end_date": datetime.date(2025, 10, 4),
                },
                None,
            ),
            {
                "begin_date": datetime.date(2015, 6, 8),
                "end_date": datetime.date(2025, 10, 4),
                "date": datetime.date(2020, 5, 20),
            },
        ),
        (
            ({"submission_day": 10, "submission_year": 2020, "submission_month": "Feb"}, "submission"),
            {
                "begin_date": datetime.date(2020, 2, 10),
                "end_date": datetime.date(2020, 2, 10),
                "submission_date": datetime.date(2020, 2, 10),
                "date": datetime.date(2020, 2, 10),
            },
        ),
        (
            ({"year": 2020, "submission_year": 2019}, "submission"),
            {"begin_date": datetime.date(2019, 1, 1), "end_date": datetime.date(2019, 12, 31)},
        ),
    ],
)
def test_get_dates(input, expected):
    actual = get_dates(input[0], input[1])
    assert actual == expected


@pytest.mark.parametrize("year,month,expected", [(2020, 2, 29), (2020, "Feb", 29)])
def test_last_day(year, month, expected):
    assert last_day(year, month) == expected


@pytest.mark.parametrize(
    "thing,expected",
    [
        ({"begin_date": "2020-01-01", "end_date": "2020-12-31"}, False),
        ({"begin_date": "2019-01-01", "end_date": "2020-12-31"}, True),
        ({"begin_date": "2019-01-01"}, True),
        ({"begin_year": 2018}, True),
        ({"begin_year": 2019}, True),
        ({"begin_year": 2020}, False),
        ({"begin_year": 2019, "begin_month": "Apr"}, True),
        ({"begin_year": 2019, "begin_month": "Jun"}, True),
        ({"begin_year": 2019, "begin_month": "Jul"}, False),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 14}, True),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 15}, True),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 16}, False),
        ({"year": 2018}, False),
        ({"year": 2019}, True),
        ({"year": 2020}, False),
        ({"year": 2019, "month": "Apr"}, False),
        ({"year": 2019, "month": "Jun"}, True),
        ({"year": 2019, "month": "Jul"}, False),
        ({"year": 2019, "month": "Jun", "day": 14}, False),
        ({"year": 2019, "month": "Jun", "day": 15}, True),
        ({"year": 2019, "month": "Jun", "day": 16}, False),
    ],
)
def test_is_current(thing, expected, now=TEST_DATE):
    assert is_current(thing, now=now) == expected


@pytest.mark.parametrize(
    "thing,expected",
    [
        ({"begin_date": "2020-01-01", "end_date": "2020-12-31"}, False),
        ({"begin_date": "2019-01-01", "end_date": "2020-12-31"}, True),
        ({"begin_date": "2019-01-01"}, True),
        ({"begin_year": 2018}, True),
        ({"begin_year": 2019}, True),
        ({"begin_year": 2020}, False),
        ({"begin_year": 2019, "begin_month": "Apr"}, True),
        ({"begin_year": 2019, "begin_month": "Jun"}, True),
        ({"begin_year": 2019, "begin_month": "Jul"}, False),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 14}, True),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 15}, True),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 16}, False),
        ({"year": 2018}, True),
        ({"year": 2019}, True),
        ({"year": 2020}, False),
        ({"year": 2019, "month": "Apr"}, True),
        ({"year": 2019, "month": "Jun"}, True),
        ({"year": 2019, "month": "Jul"}, False),
        ({"year": 2019, "month": "Jun", "day": 14}, True),
        ({"year": 2019, "month": "Jun", "day": 15}, True),
        ({"year": 2019, "month": "Jun", "day": 16}, False),
    ],
)
def test_has_started(thing, expected, now=TEST_DATE):
    assert has_started(thing, now=now) == expected


@pytest.mark.parametrize(
    "thing,expected",
    [
        ({"begin_date": "2020-01-01", "end_date": "2020-12-31"}, False),
        ({"begin_date": "2019-01-01", "end_date": "2019-06-15"}, False),
        ({"begin_date": "2019-01-01"}, False),
        ({"begin_year": 2018}, False),
        ({"begin_year": 2019}, False),
        ({"begin_year": 2020}, False),
        ({"begin_year": 2019, "begin_month": "Apr"}, False),
        ({"begin_year": 2019, "begin_month": "Jun"}, False),
        ({"begin_year": 2019, "begin_month": "Jul"}, False),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 14}, False),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 15}, False),
        ({"begin_year": 2019, "begin_month": "Jun", "begin_day": 16}, False),
        ({"year": 2018}, True),
        ({"year": 2019}, False),
        ({"year": 2020}, False),
        ({"year": 2019, "month": "Apr"}, True),
        ({"year": 2019, "month": "Jun"}, False),
        ({"year": 2019, "month": "Jul"}, False),
        ({"year": 2019, "month": "Jun", "day": 14}, True),
        ({"year": 2019, "month": "Jun", "day": 15}, False),
        ({"year": 2019, "month": "Jun", "day": 16}, False),
    ],
)
def test_has_finished(thing, expected, now=TEST_DATE):
    assert has_finished(thing, now=now) == expected


@pytest.mark.parametrize(
    "thing,expected",
    [
        ({"year": 2019, "month": "Jun", "day": 14}, True),
        ({"year": 2019, "month": "Jun", "day": 15}, False),
        ({"year": 2019, "month": "Jun", "day": 16}, False),
        ({"date": "2019-04-15"}, True),
        ({"date": "2019-08-10"}, False),
    ],
)
def test_is_before(thing, expected, now=TEST_DATE):
    assert is_before(thing, now=now) == expected


@pytest.mark.parametrize(
    "thing,expected",
    [
        ({"year": 2019, "month": "Jun", "day": 14}, False),
        ({"year": 2019, "month": "Jun", "day": 15}, False),
        ({"year": 2019, "month": "Jun", "day": 16}, True),
        ({"date": "2019-04-15"}, False),
        ({"date": "2019-08-10"}, True),
    ],
)
def test_is_after(thing, expected, now=TEST_DATE):
    assert is_after(thing, now=now) == expected


@pytest.mark.parametrize(
    "thing,expected",
    [
        ({"year": 2019, "month": "Jun", "day": 14}, False),
        ({"year": 2019, "month": "Jan", "day": 15}, True),
        ({"year": 2019, "month": "Jan", "day": 2}, True),
        ({"date": "2019-04-15"}, False),
        ({"date": "2019-02-03"}, True),
    ],
)
def test_is_between(thing, expected, start=TEST_START_DATE, end=TEST_END_DATE):
    assert is_between(thing, start=start, end=end) == expected



"""Builder Base Classes"""

import os
from glob import glob
from itertools import groupby

from jinja2 import Environment, FileSystemLoader
from xonsh.lib import subprocess

from regolith.sorters import category_val, date_key, doc_date_key, level_val
from regolith.tools import LATEX_OPTS, date_to_rfc822, gets, latex_safe, latex_safe_url, month_and_year, rfc822now


class HelperBase(object):
    """Base class for helpers"""

    def __init__(self, rc):
        self.rc = rc
        self.bldir = os.path.join(rc.builddir, self.btype)
        # allow subclasses to override
        if not hasattr(self, "env"):
            self.env = Environment(
                loader=FileSystemLoader(
                    [
                        "templates",
                        os.path.join(
                            os.path.dirname(os.path.dirname(__file__)),
                            "templates",
                        ),
                    ]
                )
            )
        self.gtx = {}
        self.construct_global_ctx()
        self.cmds = []

    def construct_global_ctx(self):
        """Constructs the global context"""
        gtx = self.gtx
        gtx["len"] = len
        gtx["True"] = True
        gtx["False"] = False
        gtx["None"] = None
        gtx["sorted"] = sorted
        gtx["groupby"] = groupby
        gtx["gets"] = gets
        gtx["date_key"] = date_key
        gtx["doc_date_key"] = doc_date_key
        gtx["level_val"] = level_val
        gtx["category_val"] = category_val
        gtx["rfc822now"] = rfc822now
        gtx["date_to_rfc822"] = date_to_rfc822

    def render(self, tname, fname, **kwargs):
        """Render the template into a file using the kwargs and global context

        Parameters
        ----------
        tname : str
            Template name
        fname : str
            Resulting file name
        kwargs : dict
            Additional kwargs to the renderer
        """
        template = self.env.get_template(tname)
        ctx = dict(self.gtx)
        ctx.update(kwargs)
        ctx["rc"] = ctx.get("rc", self.rc)
        ctx["static"] = ctx.get("static", os.path.relpath("static", os.path.dirname(fname)))
        ctx["root"] = ctx.get("root", os.path.relpath("/", os.path.dirname(fname)))
        result = template.render(ctx)
        with open(os.path.join(self.bldir, fname), "wt", encoding="utf-8") as f:
            f.write(result)

    def hlp(self):
        """run the helper, note this runs any of the commands
        listed in ``self.cmds``"""
        for cmd in self.cmds:
            getattr(self, cmd)()


class SoutHelperBase(HelperBase):
    """Base class for builders that just print to sout"""

    def __init__(self, rc):
        super().__init__(rc)
        self.cmds = ["sout"]


class DbHelperBase(HelperBase):
    """Base class for builders that update databases"""

    def __init__(self, rc):
        super().__init__(rc)
        self.cmds = ["db_updater"]


class LatexHelperBase(HelperBase):
    """Base class for Latex builders"""

    def __init__(self, rc):
        super().__init__(rc)
        self.cmds = ["latex", "clean"]

    #        if HAVE_BIBTEX_PARSER:
    #            self.bibdb = BibDatabase()
    #            self.bibwriter = BibTexWriter()

    def construct_global_ctx(self):
        super().construct_global_ctx()
        gtx = self.gtx
        gtx["month_and_year"] = month_and_year
        gtx["latex_safe"] = latex_safe
        gtx["latex_safe_url"] = latex_safe_url

    def run(self, cmd):
        """Run command in build dir"""
        subprocess.run(cmd, cwd=self.bldir, check=True)

    def pdf(self, base):
        """Compiles latex files to PDF"""
        if self.rc.pdf:
            self.run(["latex"] + LATEX_OPTS + [base + ".tex"])
            self.run(["bibtex"] + [base + ".aux"])
            self.run(["latex"] + LATEX_OPTS + [base + ".tex"])
            if os.name == "nt":
                self.run(["pdflatex"] + LATEX_OPTS + [base + ".tex"])
            else:
                self.run(["latex"] + LATEX_OPTS + [base + ".tex"])
                self.run(["dvipdf", base])

    def clean(self):
        """Remove files created by latex"""
        postfixes = [
            "*.dvi",
            "*.toc",
            "*.aux",
            "*.out",
            "*.log",
            "*.bbl",
            "*.blg",
            "*.log",
            "*.spl",
            "*~",
            "*.spl",
            "*.run.xml",
            "*-blx.bib",
        ]
        to_rm = []
        for pst in postfixes:
            to_rm += glob(os.path.join(self.bldir, pst))
        for f in set(to_rm):
            os.remove(f)



"""Builder for websites."""

import os
import shutil

from regolith.builders.basebuilder import BuilderBase
from regolith.dates import get_dates
from regolith.fsclient import _id_key
from regolith.sorters import ene_date_key, position_key
from regolith.tools import (
    all_docs_from_collection,
    dereference_institution,
    document_by_value,
    filter_projects,
    filter_publications,
    make_bibtex_file,
)


class HtmlBuilder(BuilderBase):
    """Build HTML files for website"""

    btype = "html"

    def __init__(self, rc):
        super().__init__(rc)
        # TODO: get this from the RC
        self.cmds = [
            "root_index",
            "people",
            "projects",
            "blog",
            "jobs",
            "abstracts",
            "nojekyll",
            "cname",
            "finish",
        ]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["jobs"] = list(all_docs_from_collection(rc.client, "jobs"))
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["abstracts"] = list(all_docs_from_collection(rc.client, "abstracts"))
        gtx["group"] = document_by_value(all_docs_from_collection(rc.client, "groups"), "name", rc.groupname)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["institutions"] = sorted(all_docs_from_collection(rc.client, "institutions"), key=_id_key)

    def finish(self):
        """Move files over to their destination and remove them from the
        source"""
        # static
        stsrc = os.path.join(getattr(self.rc, "static_source", "templates"), "static")
        stdst = os.path.join(self.bldir, "static")
        if os.path.isdir(stdst):
            shutil.rmtree(stdst)
        if os.path.isdir(stsrc):
            shutil.copytree(stsrc, stdst)

    def root_index(self):
        """Render root index"""
        self.render("root_index.html", "index.html", title="Home")
        make_bibtex_file(
            list(all_docs_from_collection(self.rc.client, "citations")),
            pid="group",
            person_dir=self.bldir,
        )

    def people(self):
        """Render people, former members, and each person"""
        rc = self.rc
        peeps_dir = os.path.join(self.bldir, "people")
        former_peeps_dir = os.path.join(self.bldir, "former")
        os.makedirs(peeps_dir, exist_ok=True)
        os.makedirs(former_peeps_dir, exist_ok=True)
        peeps = self.gtx["people"]
        for p in peeps:
            names = frozenset(p.get("aka", []) + [p["name"]])
            pubs = filter_publications(
                all_docs_from_collection(rc.client, "citations"),
                names,
                reverse=True,
                bold=False,
            )

            bibfile = make_bibtex_file(pubs, pid=p["_id"], person_dir=peeps_dir)
            emps = p.get("employment", [])
            emps = [em for em in emps if not em.get("not_in_cv", False)]
            for e in emps:
                e["position"] = e.get("position_full", e.get("position").title())
            ene = emps + p.get("education", [])
            ene.sort(key=ene_date_key, reverse=True)
            for e in ene:
                dereference_institution(e, all_docs_from_collection(rc.client, "institutions"))
            projs = filter_projects(all_docs_from_collection(rc.client, "projects"), names)
            for serve in p.get("service", []):
                serve_dates = get_dates(serve)
                date = serve_dates.get("date")
                if not date:
                    date = serve_dates.get("end_date")
                if not date:
                    date = serve_dates.get("begin_date")
                serve["year"] = date.year
                serve["month"] = date.month
            sns = p.get("service", [])
            sns.sort(key=ene_date_key, reverse=True)
            p["service"] = sns
            self.render(
                "person.html",
                os.path.join("people", p["_id"] + ".html"),
                p=p,
                title=p.get("name", ""),
                pubs=pubs,
                names=names,
                bibfile=bibfile,
                education_and_employment=ene,
                projects=projs,
            )
        self.render("people.html", os.path.join("people", "index.html"), title="People")

        self.render(
            "former.html",
            os.path.join("former", "index.html"),
            title="Former Members",
        )

    def projects(self):
        """Render projects"""
        rc = self.rc
        projs = all_docs_from_collection(rc.client, "projects")
        self.render("projects.html", "projects.html", title="Projects", projects=projs)

    def blog(self):
        """Render the blog and rss"""
        rc = self.rc
        blog_dir = os.path.join(self.bldir, "blog")
        os.makedirs(blog_dir, exist_ok=True)
        posts = list(all_docs_from_collection(rc.client, "blog"))
        posts.sort(key=ene_date_key, reverse=True)
        for post in posts:
            self.render(
                "blog_post.html",
                os.path.join("blog", post["_id"] + ".html"),
                post=post,
                title=post["title"],
            )
        self.render(
            "blog_index.html",
            os.path.join("blog", "index.html"),
            title="Blog",
            posts=posts,
        )
        self.render("rss.xml", os.path.join("blog", "rss.xml"), items=posts)

    def jobs(self):
        """Render the jobs and each job"""
        jobs_dir = os.path.join(self.bldir, "jobs")
        os.makedirs(jobs_dir, exist_ok=True)
        for job in self.gtx["jobs"]:
            self.render(
                "job.html",
                os.path.join("jobs", job["_id"] + ".html"),
                job=job,
                title="{0} ({1})".format(job["title"], job["_id"]),
            )
        self.render("jobs.html", os.path.join("jobs", "index.html"), title="Jobs")

    def abstracts(self):
        """Render each abstract"""
        abs_dir = os.path.join(self.bldir, "abstracts")
        os.makedirs(abs_dir, exist_ok=True)
        for ab in self.gtx["abstracts"]:
            self.render(
                "abstract.html",
                os.path.join("abstracts", ab["_id"] + ".html"),
                abstract=ab,
                title="{0} {1} - {2}".format(ab["firstname"], ab["lastname"], ab["title"]),
            )

    def nojekyll(self):
        """Touches a nojekyll file in the build dir"""
        with open(os.path.join(self.bldir, ".nojekyll"), "a+"):
            pass

    def cname(self):
        """Add CNAME"""
        rc = self.rc
        if not hasattr(rc, "cname"):
            return
        with open(os.path.join(self.bldir, "CNAME"), "w", encoding="utf-8") as f:
            f.write(rc.cname)



#!/usr/bin/env python
##############################################################################
#
# regolith         by DANSE Diffraction group
#                   Simon J. L. Billinge
#                   (c) 2011 The Trustees of Columbia University
#                   in the City of New York.  All rights reserved.
#
# File coded by:    Pavol Juhas
#
# See AUTHORS.txt for a list of people who contributed.
# See LICENSE_DANSE.txt for license information.
#
##############################################################################

"""
Definition of __version__, __date__, __timestamp__, __git_commit__.

Notes
-----
Variable `__gitsha__` is deprecated as of version 3.0.
Use `__git_commit__` instead.
"""

__all__ = ["__date__", "__git_commit__", "__timestamp__", "__version__"]

import os.path
from importlib.resources import as_file, files

# obtain version information from the version.cfg file
cp = dict(version="", date="", commit="", timestamp="0")
if __package__ is not None:
    ref = files(__package__) / "version.cfg"
    with as_file(ref) as fcfg:
        if not os.path.isfile(fcfg):  # pragma: no cover
            from warnings import warn

            warn("Package metadata not found.")
            fcfg = os.devnull
        with open(fcfg) as fp:
            kwords = [
                [w.strip() for w in line.split(" = ", 1)] for line in fp if line[:1].isalpha() and " = " in line
            ]
        assert all(w[0] in cp for w in kwords), "received unrecognized keyword"
        cp.update(kwords)
    del kwords

__version__ = cp["version"]
__date__ = cp["date"]
__git_commit__ = cp["commit"]
__timestamp__ = int(cp["timestamp"])

# TODO remove deprecated __gitsha__ in version 3.1.
__gitsha__ = __git_commit__

del cp

# End of file



"""Helper for listing upcoming (and past) grants."""

import datetime as dt

import dateutil.parser as date_parser
from gooey import GooeyParser

from regolith.dates import get_dates, is_current
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.tools import (
    all_docs_from_collection,
    collection_str,
    get_pi_id,
    key_value_pair_filter,
    merge_collections_superior,
)

TARGET_COLL = "grants"
HELPER_TARGET = "l_grants"
BLACKLIST = ["they_pay", "collgf", "physmatch", "ta", "chemmatch", "summer@seas"]


def subparser(subpi):
    date_kwargs = {}
    if isinstance(subpi, GooeyParser):
        date_kwargs["widget"] = "DateChooser"

    subpi.add_argument("-c", "--current", action="store_true", help="outputs only the current grants")
    subpi.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="if set, additional information will be printed about each grant",
    )
    subpi.add_argument(
        "-r",
        "--reveal-hidden",
        action="store_true",
        help="if set, outputs also hidden grants such as TA, " "matches etc.",
    )
    subpi.add_argument("-f", "--filter", nargs="+", help="Search this collection by giving key element pairs")
    subpi.add_argument(
        "-k",
        "--keys",
        nargs="+",
        help="Specify what keys to return values from when when running "
        "--filter. If no argument is given the default is just the id.",
    )
    subpi.add_argument("-d", "--date", help="Filter grants by a date.  Mostly used for testing.", **date_kwargs)
    return subpi


class GrantsListerHelper(SoutHelperBase):
    """Helper for listing upcoming (and past) grants."""

    # btype must be the same as helper target in helper.py
    btype = HELPER_TARGET
    needed_colls = [f"{TARGET_COLL}", "proposals"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        colls = [
            sorted(all_docs_from_collection(rc.client, collname), key=_id_key) for collname in self.needed_colls
        ]
        for db, coll in zip(self.needed_colls, colls):
            gtx[db] = coll
        gtx["grants"] = merge_collections_superior(gtx["proposals"], gtx["grants"], "proposal_id")
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        if rc.filter:
            collection = key_value_pair_filter(self.gtx["grants"], rc.filter)
        else:
            collection = self.gtx["grants"]
        grants = []
        if rc.date:
            desired_date = date_parser.parse(rc.date).date()
        else:
            desired_date = dt.date.today()
        for grant in collection:
            if rc.current and not is_current(grant, now=desired_date):
                continue
            if not rc.reveal_hidden:
                if grant.get("alias") not in BLACKLIST:
                    grants.append(grant)
            else:
                grants.append(grant)

        if rc.keys:
            results = collection_str(grants, rc.keys)
            print(results, end="")
            return
        for g in grants:
            if g.get("admin") is None:
                g["admin"] = "missing"
            g["admin"] = g.get("admin").casefold()
        grants.sort(key=lambda x: x["admin"])
        admins = list(set([g.get("admin") for g in grants]))
        for admin in admins:
            print(f"\nAdministered by: {admin}")
            sub_grants = [grant for grant in grants if grant.get("admin").strip() == admin.strip()]
            sub_grants.sort(key=lambda k: get_dates(k).get("end_date"), reverse=True)
            for g in sub_grants:
                print(
                    f"  {g.get('alias', '').ljust(15)}\t awardnr: {g.get('awardnr', '').ljust(15)}\t "
                    f"acctn: {g.get('account', 'n/a').ljust(20)}\t {get_dates(g).get('begin_date')} "
                    f"to {get_dates(g).get('end_date')}"
                )
                if rc.verbose:
                    funds_entries = g.get("funds_available")
                    if funds_entries:
                        funds_entries.sort(
                            key=lambda k: get_dates(k).get("date", get_dates(k).get("end_date")), reverse=True
                        )
                        if funds_entries[0].get("funds_available"):
                            print(
                                f"      funds available: ${funds_entries[0].get('funds_available'):,.0f} on "
                                f"{get_dates(funds_entries[0]).get('date').isoformat()}"
                            )
        return



"""API for accessing the metadata and file storage"""

import copy

from regolith.database import dump_database, open_dbs
from regolith.runcontrol import DEFAULT_RC, filter_databases, load_rcfile
from regolith.storage import push, store_client


def load_db(rc_file="regolithrc.json"):
    """Create a Broker instance from an rc file"""
    rc = copy.copy(DEFAULT_RC)
    rc._update(load_rcfile(rc_file))
    filter_databases(rc)
    return Broker(rc)


class Broker:
    """Interface to the database and file storage systems

    Examples
    --------

    >>> # Load the db
    >>> db = Broker.from_rc()
    >>> # Get a docment from the broker
    >>> ergs =db['group']['ergs']
    >>> # Store a file
    >>> db.add_file(ergs, 'myfile', '/path/to/file/hello.txt')
    >>> # Get a file from the store
    >>> path = db.get_file_path(ergs, 'myfile')
    """

    def __init__(self, rc=DEFAULT_RC):
        self.rc = rc
        # TODO: Lazy load these
        with store_client(rc) as sclient:
            self.store = sclient
        rc.client = open_dbs(rc)
        self._dbs = rc.client.dbs
        self.md = rc.client.chained_db
        self.db_client = rc.client

    def add_file(self, document, name, filepath):
        """Add a file to a document in a collection.

        Parameters
        ----------
        document : dict
            The document to add the file to
        name : str
            Name of the reference to the file
        filepath : str
            Location of the file on local disk
        """
        output_path = self.store.copydoc(filepath)
        if "files" not in document:
            document["files"] = {}
        document["files"][name] = output_path
        for db in self.rc.databases:
            dump_database(db, self.db_client, self.rc)
        push(self.store.store, self.store.path)

    @classmethod
    def from_rc(cls, rc_file="regolithrc.json"):
        """Return a Broker instance"""
        return load_db(rc_file)

    def get_file_path(self, document, name):
        """Get a file from the file storage associated with the document and
        name

        Parameters
        ----------
        document : dict
            The document which stores the reference to the file
        name : str
            The name of the file stored (note that this can be different from
            the filename itself)

        Returns
        -------
        path : str or None
            The file path, if not in the storage None
        """
        if "files" in document:
            return self.store.retrieve(document["files"][name])
        else:
            return None

    def __getitem__(self, item):
        return self.md[item]



from regolith.schemas import _update_dict_target, insert_alloweds


def test_update_dict_target():
    doc = {
        "email": {
            "description": "contact email for the author.",
            "required": True,
            "deeper": {"eallowed": "TEST"},
        },
        "_id": {
            "description": "Unique identifier for submission. "
            "This generally includes the author name and part of the title.",
            "required": True,
            "type": "string",
        },
        "coauthors": {"description": "names of coauthors", "required": False, "eallowed": "TEST2"},
        "test_repeated": {"description": "names of coauthors", "required": False, "eallowed": "TEST2"},
    }
    expected = {
        "_id": {
            "description": "Unique identifier for submission. "
            "This generally includes the author name and part of the title.",
            "required": True,
            "type": "string",
        },
        "coauthors": {"description": "names of coauthors", "required": False, "eallowed": "string"},
        "test_repeated": {"description": "names of coauthors", "required": False, "eallowed": "string"},
        "email": {
            "description": "contact email for the author.",
            "required": True,
            "deeper": {"eallowed": ["string", "float"]},
        },
    }
    first_cut = _update_dict_target(doc, {"eallowed": "TEST"}, ["string", "float"])
    actual = _update_dict_target(first_cut, {"eallowed": "TEST2"}, "string")
    assert actual == expected


def test_insert_alloweds():
    alloweds = {"TEST": ["string", "float"], "TEST2": "string"}
    doc = {
        "email": {
            "description": "contact email for the author.",
            "required": True,
            "deeper": {"eallowed": "TEST"},
        },
        "_id": {
            "description": "Unique identifier for submission. "
            "This generally includes the author name and part of the title.",
            "required": True,
            "type": "string",
        },
        "coauthors": {"description": "names of coauthors", "required": False, "eallowed": "TEST2"},
        "test_repeated": {"description": "names of coauthors", "required": False, "eallowed": "TEST2"},
    }
    expected = {
        "_id": {
            "description": "Unique identifier for submission. "
            "This generally includes the author name and part of the title.",
            "required": True,
            "type": "string",
        },
        "coauthors": {"description": "names of coauthors", "required": False, "eallowed": "string"},
        "test_repeated": {"description": "names of coauthors", "required": False, "eallowed": "string"},
        "email": {
            "description": "contact email for the author.",
            "required": True,
            "deeper": {"eallowed": ["string", "float"]},
        },
    }
    actual = insert_alloweds(doc, alloweds, "eallowed")
    assert actual == expected



"""Validators and converters for regolith input."""

import os
from getpass import getpass

from regolith.tools import string_types


def noop(x):
    """Does nothing, just returns the input."""
    return x


def is_int(x):
    """Tests if something is an integer"""
    return isinstance(x, int)


def always_true(x):
    """Returns True"""
    return True


def always_false(x):
    """Returns False"""
    return False


def is_bool(x):
    """Tests if something is a boolean"""
    return isinstance(x, bool)


def is_string(x):
    """Tests if something is a string"""
    return isinstance(x, string_types)


_FALSES = frozenset(["", "0", "n", "f", "no", "none", "false"])


def to_bool(x):
    """ "Converts to a boolean in a semantically meaningful way."""
    if isinstance(x, bool):
        return x
    elif isinstance(x, string_types):
        return False if x.lower() in _FALSES else True
    else:
        return bool(x)


def ensure_string(x):
    """Returns a string if x is not a string, and x if it already is."""
    if isinstance(x, string_types):
        return x
    else:
        return str(x)


def ensure_database(db):
    db["name"] = ensure_string(db["name"])
    db["url"] = ensure_string(db["url"])
    db["path"] = ensure_string(db["path"])
    db["public"] = to_bool(db.get("public", True))
    return db


def ensure_databases(dbs):
    """Ensures each dataset in a list of databases"""
    return list(map(ensure_database, dbs))


def ensure_store(store):
    store["name"] = ensure_string(store["name"])
    store["url"] = ensure_string(store["url"])
    store["path"] = ensure_string(store.get("path", None) or "")
    store["public"] = to_bool(store.get("public", True))
    return store


def ensure_stores(stores):
    """Ensures each store in a list of stores"""
    return list(map(ensure_store, stores))


def ensure_email(email):
    """Ensures the email top-level key is well formed."""
    email["url"] = ensure_string(email["url"])
    if "cred" in email:
        email["cred"] = ensure_string(email["cred"])
    else:
        email["cred"] = email["url"] + ".cred"
    if not os.path.isfile(email["cred"]):
        user = input("Email address for " + email["url"] + ": ")
        password = getpass()
        s = user + "\n" + password
        with open(email["cred"], "w") as f:
            f.write(s)
    with open(email["cred"], encoding="utf-8") as f:
        email["from"] = f.readline().strip()
        email["password"] = f.readline().strip()
    email["user"] = email["from"].partition("@")[0]
    email["port"] = int(email.get("port", 0))
    email["verbosity"] = int(email.get("verbosity", 0))
    email["tls"] = to_bool(email.get("tls", False))
    return email


DEFAULT_VALIDATORS = {
    "backend": (is_string, ensure_string),
    "builddir": (is_string, ensure_string),
    "databases": (always_false, ensure_databases),
    "stores": (always_false, ensure_stores),
    "email": (always_false, ensure_email),
}



"""Builder for Current and Pending Reports."""

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.fsclient import _id_key
from regolith.tools import all_docs_from_collection


class PostdocadBuilder(LatexBuilderBase):
    """Build current and pending report from database entries"""

    btype = "postdocads"
    needed_colls = ["postdocads"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["postdocads"] = sorted(all_docs_from_collection(rc.client, "postdocads"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        """Render latex template"""
        for ads in self.gtx["postdocads"]:
            goals = ads["projectGoals"]
            positionOn = ads["positionOn"]
            projectTasks = ads["projectTasks"]
            requiredExperience = ads["requiredExperience"]
            additionalDesiredExperience = ads["additionalDesiredExperience"]
            startDate = ads["startDate"]
            thirdYear = ads["thirdyear"]
            k = ads["_id"]
            applicationDeadline = ads["applicationDeadline"]
            outname = "{}".format(k)
            self.render(
                "postdocad.tex",
                outname + ".tex",
                projectGoals=goals,
                positionOn=positionOn,
                projectTasks=projectTasks,
                requiredExperience=requiredExperience,
                additionalDesiredExperience=additionalDesiredExperience,
                startDate=startDate,
                thirdYear=thirdYear,
                k=k,
                applicationDeadline=applicationDeadline,
            )
            self.pdf("test")



"""Helper for adding a proposal to the proposals.yml collection."""

import datetime as dt
from math import floor

import dateutil.parser as date_parser
from dateutil.relativedelta import relativedelta
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.tools import all_docs_from_collection, get_pi_id

TARGET_COLL = "proposals"


def subparser(subpi):
    amount_kwargs = {}
    notes_kwargs = {}
    date_kwargs = {}
    dropdown_kwargs = {}
    int_kwargs = {}
    if isinstance(subpi, GooeyParser):
        amount_kwargs["widget"] = "DecimalField"
        amount_kwargs["gooey_options"] = {"min": 0.00, "max": 1000000.00, "increment": 10.00, "precision": 2}
        notes_kwargs["widget"] = "Textarea"
        date_kwargs["widget"] = "DateChooser"
        dropdown_kwargs["widget"] = "Dropdown"
        dropdown_kwargs["choices"] = [True, False]
        int_kwargs["widget"] = "IntegerField"

    # Do not delete --database arg
    subpi.add_argument(
        "name",
        help="A short but unique name for the proposal",
    )
    subpi.add_argument(
        "amount",
        help="value of award",
    )
    subpi.add_argument("title", help="Actual title of the proposal")
    subpi.add_argument("--begin-date", help="The begin date for the proposed grant ", **date_kwargs)
    subpi.add_argument(
        "--duration",
        help="The duration for the proposed grant in months. " "Please enter either an end date or a duration. ",
        **int_kwargs,
    )
    subpi.add_argument(
        "--end-date",
        help="The end date for the proposed grant." " Please enter either an " "end date or a duration",
        **date_kwargs,
    )
    subpi.add_argument("--due-date", help="The due date for the proposal", **date_kwargs)
    subpi.add_argument("-a", "--authors", nargs="+", help="Other investigator names", default=[])
    subpi.add_argument(
        "--not-cpp",
        action="store_true",
        help="Check if the proposal should not appear in the " "current and pending support form",
    )
    subpi.add_argument(
        "--other-agencies",
        help="Other agencies to which the proposal has been " "submitted. Defaults to None",
        default="None",
    )
    subpi.add_argument(
        "-i", "--institution", help="The institution where the work will primarily " "be carried out", default=""
    )
    subpi.add_argument(
        "-p", "--pi", help="ID of principal investigator. Defaults to " "group pi id in regolithrc.json"
    )
    subpi.add_argument(
        "--months-academic",
        help="Number of working months in the academic year "
        "to be stated on the current and pending form. "
        "Defaults to 0",
        default=0,
        **int_kwargs,
    )
    subpi.add_argument(
        "--months-summer",
        help="Number of working months in the summer to be "
        "stated on the current and pending form. "
        "Defaults to 0",
        default=0,
        **int_kwargs,
    )
    subpi.add_argument(
        "-s",
        "--scope",
        help="Scope of project and statement of any overlaps " "with other current and pending grants",
        default="",
    )
    subpi.add_argument("-f", "--funder", help="Agency where the proposal is being submitted", default="")
    subpi.add_argument("-n", "--notes", nargs="+", help="Anything to note", default=[], **notes_kwargs)
    subpi.add_argument(
        "-c", "--currency", help="Currency in which amount is specified. Defaults to USD", default="USD"
    )
    subpi.add_argument(
        "--database",
        help="The database that will be updated.  Defaults to " "first database in the regolithrc.json file.",
    )
    subpi.add_argument("--date", help="The date that will be used for testing.", **date_kwargs)
    return subpi


class ProposalAdderHelper(DbHelperBase):
    """Helper for adding a proposal to the proposals collection.

    A proposal is a dictionary object describing a research or
     project proposal submitted by the group.
    """

    # btype must be the same as helper target in helper.py
    btype = "a_proposal"
    needed_colls = [f"{TARGET_COLL}", "people", "groups"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        rc.pi_id = get_pi_id(rc)
        rc.coll = f"{TARGET_COLL}"
        if not rc.database:
            rc.database = rc.databases[0]["name"]
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        if rc.date:
            now = date_parser.parse(rc.date).date()
        else:
            now = dt.date.today()
        key = f"{str(now.year)[2:]}_{''.join(rc.name.casefold().split()).strip()}"

        coll = self.gtx[rc.coll]
        pdocl = list(filter(lambda doc: doc["_id"] == key, coll))
        if len(pdocl) > 0:
            raise RuntimeError("This entry appears to already exist in the collection")
        else:
            pdoc = {}
        pdoc.update({"_id": key, "amount": float(rc.amount)})
        pdoc.update({"authors": rc.authors})
        if rc.begin_date:
            begin_date = date_parser.parse(rc.begin_date).date()
            if rc.end_date:
                end_date = date_parser.parse(rc.end_date).date()
                if begin_date > end_date:
                    raise ValueError("begin_date can not be after end_date")
                if rc.duration:
                    expected_duration = relativedelta(end_date, begin_date)
                    input_duration = float(rc.duration)
                    input_years = floor(input_duration / 12)
                    input_months = floor(input_duration) % 12
                    input_days = (input_duration % 1) * 30.5
                    # we take 30.5 as an average of 30 and 31 day months
                    if (
                        expected_duration.years != input_years
                        or expected_duration.months != input_months
                        or abs(expected_duration.days - input_days) > 3.05
                    ):
                        # assuming that the user inputs the correct duration up to 1 decimal place
                        # so the maximum allowed difference is 0.1*30.5
                        raise ValueError("ERROR: please rerun specifying a duration OR an end-date but not both")
            pdoc.update({"begin_date": begin_date})
        else:
            pdoc.update({"begin_date": "tbd"})
        cpp_info = {}
        if rc.not_cpp:
            cpp_info["cppflag"] = False
        else:
            cpp_info["cppflag"] = True
        cpp_info["other_agencies_submitted"] = rc.other_agencies
        cpp_info["institution"] = rc.institution
        cpp_info["person_months_academic"] = float(rc.months_academic)
        cpp_info["person_months_summer"] = float(rc.months_summer)
        cpp_info["project_scope"] = rc.scope
        pdoc.update({"cpp_info": cpp_info})
        pdoc.update({"currency": rc.currency})
        if rc.due_date:
            pdoc.update({"due_date": date_parser.parse(rc.due_date).date()})
        else:
            pdoc.update({"due_date": "tbd"})
        if rc.end_date:
            pdoc.update({"end_date": date_parser.parse(rc.end_date).date()})
        elif rc.begin_date and rc.duration and not rc.end_date:
            begin_date = date_parser.parse(rc.begin_date).date()
            pdoc.update({"end_date": begin_date + relativedelta(days=30.5 * float(rc.duration))})
        else:
            pdoc.update({"end_date": "tbd"})
        pdoc.update({"funder": rc.funder})
        pdoc.update({"notes": rc.notes})
        if rc.pi:
            pdoc.update({"pi": rc.pi})
        else:
            pdoc.update({"pi": rc.pi_id})
        pdoc.update({"status": "inprep"})
        sample_team = {"name": "", "subaward_amount": 0}
        pdoc.update({"team": [sample_team]})
        pdoc.update({"title": rc.title})

        rc.client.insert_one(rc.database, rc.coll, pdoc)

        print(f"{key} has been added in {TARGET_COLL}")

        return



import os
from copy import copy, deepcopy

import pytest

from regolith.database import connect
from regolith.runcontrol import DEFAULT_RC, load_rcfile
from regolith.schemas import EXEMPLARS
from regolith.tools import all_docs_from_collection


def test_collection_retrieval_python(make_mixed_db):
    if make_mixed_db is False:
        pytest.skip("Mongoclient failed to start")
    else:
        repo, fs_coll, mongo_coll = make_mixed_db
    os.chdir(repo)
    rc = copy(DEFAULT_RC)
    rc._update(load_rcfile("regolithrc.json"))
    with connect(rc) as rc.client:
        fs_test_dict = dict(list(all_docs_from_collection(rc.client, "abstracts"))[0])
        mongo_test_dict = dict(list(all_docs_from_collection(rc.client, "assignments"))[0])
    fs_expected_dict = deepcopy(EXEMPLARS[fs_coll])
    mongo_expected_dict = deepcopy(EXEMPLARS[mongo_coll])
    assert fs_test_dict == fs_expected_dict
    assert mongo_test_dict == mongo_expected_dict



"""Database schemas, examples, and tools"""

import copy
import json
from pathlib import Path
from warnings import warn

from cerberus import Validator
from flatten_dict import flatten, unflatten

from .sorters import POSITION_LEVELS

SORTED_POSITION = sorted(POSITION_LEVELS.keys(), key=POSITION_LEVELS.get)
PROJECTUM_ACTIVE_STATI = ["proposed", "converged", "started"]
PROJECTUM_PAUSED_STATI = ["backburner", "paused"]
PROJECTUM_CANCELLED_STATI = ["cancelled"]
PROJECTUM_FINISHED_STATI = ["finished"]
PROJECTUM_STATI = list(
    PROJECTUM_ACTIVE_STATI + PROJECTUM_PAUSED_STATI + PROJECTUM_CANCELLED_STATI + PROJECTUM_FINISHED_STATI
)

alloweds = {
    "ACTIVITIES_TYPES": ["teaching", "research"],
    "AGENCIES": ["nsf", "doe", "other"],
    "APPOINTMENTS_TYPES": ["gra", "ss", "pd", "ug"],
    "APPOINTMENTS_STATI": ["proposed", "appointed", "finalized"],
    "COMMITTEES_TYPES": ["phdoral", "phddefense", "phdproposal", "promotion"],
    "COMMITTEES_LEVELS": ["department", "school", "university", "external"],
    "EXPENSES_STATI": ["unsubmitted", "submitted", "reimbursed", "declined"],
    "EXPENSES_TYPES": ["travel", "business"],
    "FACILITIES_TYPES": [
        "teaching",
        "research",
        "shared",
        "other",
        "teaching_wish",
        "research_wish",
    ],
    "GRANT_STATI": ["pending", "declined", "accepted", "in-prep"],
    "GRANT_ROLES": ["pi", "copi"],
    "MILESTONE_TYPES": [
        "mergedpr",
        "meeting",
        "other",
        "paper",
        "release",
        "email",
        "handin",
        "purchase",
        "approval",
        "presentation",
        "report",
        "submission",
        "decision",
        "demo",
        "skel",
    ],
    "POSITION_STATI": [
        "pi",
        "adjunct",
        "high-school",
        "undergrad",
        "ms",
        "phd",
        "postdoc",
        "visitor-supported",
        "visitor-unsupported",
        "research-associate",
    ],
    "PRESENTATION_TYPES": [
        "award",
        "colloquium",
        "contributed_oral",
        "invited",
        "keynote",
        "plenary",
        "poster",
        "seminar",
        "tutorial",
        "other",
    ],
    "PRESENTATION_STATI": [
        "in-prep",
        "submitted",
        "accepted",
        "declined",
        "cancelled",
        "postponed",
    ],
    "PROJECT_TYPES": ["ossoftware", "funded", "outreach"],
    "PROJECTUM_ACTIVE_STATI": PROJECTUM_ACTIVE_STATI,
    "PROJECTUM_PAUSED_STATI": PROJECTUM_PAUSED_STATI,
    "PROJECTUM_CANCELLED_STATI": PROJECTUM_CANCELLED_STATI,
    "PROJECTUM_FINISHED_STATI": PROJECTUM_FINISHED_STATI,
    "PROJECTUM_STATI": PROJECTUM_STATI,
    "PROPOSAL_STATI": ["pending", "declined", "accepted", "inprep", "submitted"],
    "PUBLICITY_TYPES": ["online", "article"],
    "REVIEW_STATI": [
        "invited",
        "accepted",
        "declined",
        "downloaded",
        "inprogress",
        "submitted",
        "cancelled",
    ],
    "REVIEW_RECOMMENDATIONS": ["reject", "asis", "smalledits", "diffjournal", "majoredits"],
    "SERVICE_TYPES": ["profession", "university", "school", "department"],
    "SORTED_POSITION": SORTED_POSITION,
    "TODO_STATI": ["started", "finished", "cancelled", "paused"],
    "OPTIONAL_KEYS_INSTITUTIONS": [
        "aka",
        "departments",
        "schools",
        "state",
        "street",
        "zip",
    ],
    # for status of kickoff, deliverable, milestones, and the projectum
}


def _update_dict_target(d, filter, new_value):
    flatd = flatten(d)
    for filtk, filtv in filter.items():
        for k, v in flatd.items():
            if filtk in k:
                if filtv == v:
                    flatd.update({k: new_value})
    unflatd = unflatten(flatd)
    return unflatd


def insert_alloweds(doc, alloweds, key):
    working_doc = copy.deepcopy(doc)
    for k, v in alloweds.items():
        working_doc = _update_dict_target(working_doc, {key: k}, v)
    return working_doc


def load_schemas():
    here = Path(__file__).parent
    schema_file = here / "schemas.json"
    with open(schema_file, "r", encoding="utf-8") as schema_file:
        raw_schemas = json.load(schema_file)
    schemas = insert_alloweds(raw_schemas, alloweds, "eallowed")
    return schemas


def load_exemplars():
    here = Path(__file__).parent
    exemplar_file = here / "exemplars.json"
    with open(exemplar_file, "r", encoding="utf-8") as efile:
        exemplars = json.load(efile)
    return exemplars


EXEMPLARS = load_exemplars()
SCHEMAS = load_schemas()

for s in SCHEMAS:
    SCHEMAS[s]["files"] = {
        "description": "Files associated with the document",
        # TODO: fix this since this is currently comming out a CommentedMap (+1: Yevgeny)
        # "type": "list",
        # "schema": {"type": "string"},
        "required": False,
    }


class NoDescriptionValidator(Validator):
    def _validate_description(self, description, field, value):
        """Don't validate descriptions

        The rule's arguments are validated against this schema:
        {'type': 'string'}"""
        if False:
            pass

    def _validate_eallowed(self, eallowed, field, value):
        """Test if value is in list
        The rule's arguments are validated against this schema:
        {'type': 'list'}
        """
        if value not in eallowed:
            warn(
                '"{}" is not in the preferred entries for "{}", please '
                "consider changing this entry to conform or add this to the "
                "``eallowed`` field in the schema.".format(value, field)
            )


def validate(coll, record, schemas):
    """Validate a record for a given db

    Parameters
    ----------
    coll : str
        The name of the db in question
    record : dict
        The record to be validated
    schemas : dict
        The schema to validate against

    Returns
    -------
    rtn : bool
        True is valid
    errors: dict
        The errors encountered (if any)

    """
    if coll in schemas:
        schema = copy.deepcopy(schemas[coll])
        v = NoDescriptionValidator(schema)
        return v.validate(record), v.errors
    else:
        return True, ()



import calendar
from datetime import date, timedelta

from regolith.dates import get_dates
from regolith.fsclient import _id_key
from regolith.helpers.basehelper import SoutHelperBase
from regolith.helpers.makeappointmentshelper import _future_grant
from regolith.sorters import position_key
from regolith.tools import all_docs_from_collection, fuzzy_retrieval, merge_collections_superior

SEMESTER_START_MONTH = {"fall": (9, 12), "spring": (1, 5), "summer": (6, 8)}


def subparser(subpi):
    subpi.add_argument(
        "-d",
        "--date",
        help="the date from which its current appointments will be listed, " "defaults to today's date",
    )
    subpi.add_argument(
        "-s",
        "--semester",
        help="list all the appointments for the semester of the specified date",
        action="store_true",
    )
    return subpi


class CurrentAppointmentsListerHelper(SoutHelperBase):
    """Helper for managing appointments on grants and studying the burn of grants over time."""

    # btype must be the same as helper target in helper.py
    btype = "currentappointments"
    needed_colls = ["people", "grants", "proposals"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        gtx["people"] = sorted(
            all_docs_from_collection(rc.client, "people"),
            key=position_key,
            reverse=True,
        )
        gtx["grants"] = sorted(all_docs_from_collection(rc.client, "grants"), key=_id_key)
        gtx["proposals"] = sorted(all_docs_from_collection(rc.client, "proposals"), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def sout(self):
        rc = self.rc
        since, before = None, None
        if rc.date:
            ondate = date(*[int(num) for num in rc.date.split("-")])
        else:
            ondate = date.today()

        if rc.semester:
            for v in SEMESTER_START_MONTH.values():
                if v[0] <= ondate.month <= v[1]:
                    since = date(ondate.year, v[0], 1)
                    last_day = calendar.monthrange(ondate.year, v[1])[1]
                    before = date(ondate.year, v[1], last_day)

        people = self.gtx["people"]
        jg = self.gtx["grants"]
        proposals = self.gtx["proposals"]
        grants = merge_collections_superior(proposals, jg, "proposal_id")
        _future_grant["begin_date"] = ondate
        _future_grant["end_date"] = ondate + timedelta(days=2190)
        _future_grant["budget"][0]["begin_date"] = ondate
        _future_grant["budget"][0]["end_date"] = ondate + timedelta(days=2190)
        grants.append(_future_grant)
        for person in people:
            p_appt = person.get("appointments", None)
            if p_appt:
                for _id, appt in p_appt.items():
                    grantid = appt.get("grant")
                    if not grantid:
                        print(f"No grant found in {person.get('_id')} appt {person.get('appt')}")
                    grant = fuzzy_retrieval(grants, ["name", "_id", "alias"], grantid)
                    if not grant:
                        print(f"No grant found for {grantid}")
                    try:
                        accountnr = grant.get("account", grant["alias"])
                    except Exception:
                        accountnr = ""
                    loading = appt.get("loading")
                    appt_dates = get_dates(appt)
                    bd = appt_dates.get("begin_date")
                    ed = appt_dates.get("end_date")
                    if since and before:
                        if (ed >= since) and (bd <= before):
                            print(
                                person.get("_id"),
                                grantid,
                                accountnr,
                                loading,
                                bd.strftime("%Y-%m-%d"),
                                ed.strftime("%Y-%m-%d"),
                            )
                    else:
                        if bd <= ondate <= ed:
                            print(
                                person.get("_id"),
                                grantid,
                                accountnr,
                                loading,
                                bd.strftime("%Y-%m-%d"),
                                ed.strftime("%Y-%m-%d"),
                            )



"""Builder for publication lists."""

import string
from datetime import datetime

from dateutil import parser as date_parser

from regolith.builders.basebuilder import LatexBuilderBase
from regolith.sorters import position_key
from regolith.tools import all_docs_from_collection

LATEX_OPTS = ["-halt-on-error", "-file-line-error"]


class FormalLetterBuilder(LatexBuilderBase):
    btype = "formalletter"
    needed_colls = ["formalletters"]

    def construct_global_ctx(self):
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if not rc.people:
            rc.people = ["all"]
        if isinstance(rc.people, str):
            rc.people = [rc.people]

        gtx["formalletters"] = sorted(
            all_docs_from_collection(rc.client, "formalletters"),
            key=position_key,
            reverse=True,
        )
        try:
            rc.ranks_and_roles.update(
                {
                    "lc": "Lieutenant Commander",
                    "co": "Commanding Officer",
                    "fl": "First Lieutenant",
                    "fls": "1stLt",
                }
            )
        except AttributeError:
            rc.ranks_and_roles = {
                "lc": "Lieutenant Commander",
                "co": "Commanding Officer",
                "fl": "First Lieutenant",
                "fls": "1stLt",
            }
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def latex(self):
        filestub, qualifiers = "", ""
        if self.rc.from_date:
            from_date = date_parser.parse(self.rc.from_date).date()
            filestub = filestub + "_from{}".format(from_date)
            qualifiers = qualifiers + f"in the period from {from_date}"
            if self.rc.to_date:
                to_date = date_parser.parse(self.rc.to_date).date()
                filestub = filestub + "_to{}".format(to_date)
                qualifiers = qualifiers + " to {}".format(to_date)
            else:
                to_date = None
        else:
            from_date, to_date = None, None

        for letter in self.gtx["formalletters"]:
            if letter.get("date") is type(datetime.date):
                letter_date = letter.get("date")
            else:
                letter_date = date_parser.parse(letter.get("date")).date()
            letter["date"] = letter_date.strftime("%m/%e/%y").strip("0").replace(" ", "")
            outfile_name_stub = f"{letter.get('_id')}"
            to = letter.get("to")
            if to.get("title") in self.rc.ranks_and_roles.keys():
                to["title"] = self.rc.ranks_and_roles.get(to.get("title"))
            nc_to = f"{to.get('title')} {to.get('name')} {to.get('postfix')}"
            fr = letter.get("from")
            if fr.get("title") in self.rc.ranks_and_roles.keys():
                fr["title"] = self.rc.ranks_and_roles.get(fr.get("title"))
            nc_fr = f"{fr.get('title')} {fr.get('name')} {fr.get('postfix')}"
            print(nc_fr)
            # The subject has to be upper case and have no trailing period, but what
            # if it wasn't entered in the DB like that...this ensures that it appears
            # this way in the final letter.
            letter["subject"] = letter["subject"].upper().strip(".")
            letter["encls"] = [
                f"({string.ascii_lowercase[i]}) {enc[:1].upper()+enc[1:].strip('.')}"
                for i, enc in enumerate(letter["encls"])
            ]
            letter["refs"] = [
                f"({string.ascii_lowercase[i]}) {ref[:1].upper()+ref[1:].strip('.')}"
                for i, ref in enumerate(letter["refs"])
            ]
            # ensure that all paras end in a (single) period
            letter["paras"] = [f"{para.strip('.').capitalize()}." for para in letter.get("paras", "")]
            self.render(
                "naval_correspondence.tex", outfile_name_stub + "_nc.tex", contents=letter, to=nc_to, fr=nc_fr
            )



"""Helper for updating a task in todos of todos collection."""

import datetime as dt
import math

import dateutil.parser as date_parser
from dateutil.relativedelta import relativedelta
from gooey import GooeyParser

from regolith.fsclient import _id_key
from regolith.helpers.basehelper import DbHelperBase
from regolith.schemas import PROJECTUM_ACTIVE_STATI, alloweds
from regolith.tools import (
    all_docs_from_collection,
    document_by_value,
    get_pi_id,
    key_value_pair_filter,
    print_task,
)

TARGET_COLL = "todos"
ALLOWED_IMPORTANCE = [3, 2, 1, 0]

TODO_STATI = alloweds.get("TODO_STATI")


def subparser(subpi):
    deci_kwargs = {}
    notes_kwargs = {}
    date_kwargs = {}
    int_kwargs = {}
    listbox_kwargs = {}
    if isinstance(subpi, GooeyParser):
        deci_kwargs["widget"] = "DecimalField"
        deci_kwargs["gooey_options"] = {"min": 0.0, "max": 10000.0, "increment": 1, "precision": 1}
        notes_kwargs["widget"] = "Textarea"
        date_kwargs["widget"] = "DateChooser"
        int_kwargs["widget"] = "IntegerField"
        listbox_kwargs["widget"] = "Listbox"
        listbox_kwargs["choices"] = TODO_STATI

    subpi.add_argument(
        "-i",
        "--index",
        help="Enter the index of a certain task in the enumerated list to update that task.",
        type=int,
        **int_kwargs,
    )
    subpi.add_argument(
        "-s",
        "--stati",
        nargs="+",
        help='Update tasks with specific stati"',
        default=["started"],
        **listbox_kwargs,
    )
    subpi.add_argument(
        "-f",
        "--filter",
        nargs="+",
        help="Search this collection by giving key element pairs. "
        "'-f description paper' will return tasks with description containing 'paper' ",
    )
    subpi.add_argument("-r", "--renumber", action="store_true", help="Renumber the indices.")
    subpi.add_argument(
        "-d",
        "--description",
        help=" Change the description of the to_do task. If the description has more than one "
        "word, please enclose it in quotation marks.",
    )
    subpi.add_argument("-u", "--due-date", help="Change the due date of the task.", **date_kwargs)
    subpi.add_argument(
        "-e",
        "--estimated-duration",
        help="Change the estimated duration the task will take in minutes. ",
        type=float,
        **deci_kwargs,
    )
    subpi.add_argument(
        "--deadline", help="give value 't' if due_date is a hard deadline, else 'f' if not", choices=["t", "f"]
    )
    subpi.add_argument(
        "-m", "--importance", choices=ALLOWED_IMPORTANCE, help="Change the importance of the task.", type=int
    )
    subpi.add_argument("--status", choices=TODO_STATI, help="Change the status of the task.")
    subpi.add_argument(
        "-n",
        "--notes",
        nargs="+",
        help="The new notes for this task. Each note should be enclosed " "in quotation marks.",
        **notes_kwargs,
    )
    subpi.add_argument("-t", "--tags", nargs="+", help="The new tags to add for this task.")
    subpi.add_argument("--begin-date", help="Change the begin date of the task.", **date_kwargs)
    subpi.add_argument("--end-date", help="Change the end date of the task.", **date_kwargs)
    subpi.add_argument(
        "-a",
        "--assigned-to",
        help="Filter tasks that are assigned to this user id. Default id is saved in user.json. ",
    )
    subpi.add_argument(
        "-b",
        "--assigned-by",
        nargs="?",
        const="default_id",
        help="Filter tasks that are assigned to other members by this user id. Default id is saved in user.json. ",
    )
    subpi.add_argument(
        "--date",
        help="Enter a date such that the helper can calculate how many days are left "
        "from that date to the due date. Default is today.",
        **date_kwargs,
    )

    return subpi


class TodoUpdaterHelper(DbHelperBase):
    """Helper for updating a task in todos of todos collection."""

    # btype must be the same as helper target in helper.py
    btype = "u_todo"
    needed_colls = [f"{TARGET_COLL}"]

    def construct_global_ctx(self):
        """Constructs the global context"""
        super().construct_global_ctx()
        gtx = self.gtx
        rc = self.rc
        if "groups" in self.needed_colls:
            rc.pi_id = get_pi_id(rc)

        rc.coll = f"{TARGET_COLL}"
        gtx[rc.coll] = sorted(all_docs_from_collection(rc.client, rc.coll), key=_id_key)
        gtx["all_docs_from_collection"] = all_docs_from_collection
        gtx["float"] = float
        gtx["str"] = str
        gtx["zip"] = zip

    def db_updater(self):
        rc = self.rc
        if rc.index:
            if rc.index >= 9900:
                print(
                    "WARNING: indices >= 9900 are used for milestones which "
                    "should be updated using u_milestone and not u_todo"
                )
                return
        if not rc.assigned_to:
            try:
                rc.assigned_to = rc.default_user_id
            except AttributeError:
                print(
                    "Please set default_user_id in '~/.config/regolith/user.json', or you need to enter "
                    "your group id in the command line"
                )
                return
        filterid = {"_id": rc.assigned_to}
        person = document_by_value(all_docs_from_collection(rc.client, "todos"), "_id", rc.assigned_to)
        if not person:
            raise TypeError(f"Id {rc.assigned_to} can't be found in todos collection")
        todolist = person.get("todos", [])
        if len(todolist) == 0:
            print(f"{rc.assigned_to} doesn't have todos in todos collection.")
            return
        if not rc.date:
            today = dt.date.today()
        else:
            today = date_parser.parse(rc.date).date()
        if not rc.index:
            finished_todo = 0
            for todo in todolist:
                if todo["status"] == "finished":
                    finished_todo += 1
                if isinstance(todo.get("due_date"), str):
                    todo["due_date"] = date_parser.parse(todo["due_date"]).date()
                if isinstance(todo.get("end_date"), str):
                    todo["end_date"] = date_parser.parse(todo["end_date"]).date()
                todo["days_to_due"] = (todo.get("due_date") - today).days
                todo["sort_finished"] = (todo.get("end_date", dt.date(1900, 1, 1)) - dt.date(1900, 1, 1)).days
                todo["order"] = 1 / (1 + math.exp(abs(todo["days_to_due"] - 0.5)))
            todolist = sorted(
                todolist, key=lambda k: (k["status"], k["importance"], k["order"], -k.get("duration", 10000))
            )
            todolist[:finished_todo] = sorted(todolist[:finished_todo], key=lambda k: (-k["sort_finished"]))
            index_match = {}
            if rc.renumber:
                new_index_started = 1
                new_index_finished = -1
                for todo in reversed(todolist[finished_todo:]):
                    index_match[todo["running_index"]] = new_index_started
                    new_index_started += 1
                for todo in reversed(todolist[:finished_todo]):
                    index_match[todo["running_index"]] = new_index_finished
                    new_index_finished += -1
                for i in range(0, len(rc.databases)):
                    db_name = rc.databases[i]["name"]
                    person_idx = rc.client.find_one(db_name, rc.coll, filterid)
                    if isinstance(person_idx, dict):
                        todolist_idx = person_idx.get("todos", [])
                    else:
                        continue
                    if len(todolist_idx) != 0:
                        for todo in todolist_idx:
                            index = index_match[todo["running_index"]]
                            todo["running_index"] = index
                        rc.client.update_one(
                            db_name, rc.coll, {"_id": rc.assigned_to}, {"todos": todolist_idx}, upsert=True
                        )
                        print(f"Indices in {db_name} for {rc.assigned_to} have been updated.")
                return
            if rc.assigned_by:
                if rc.assigned_by == "default_id":
                    rc.assigned_by = rc.default_user_id
                for todo in todolist[::-1]:
                    if todo.get("assigned_by") != rc.assigned_by:
                        print(todo.get("assigned_by"))
                        todolist.remove(todo)
            if rc.filter:
                todolist = key_value_pair_filter(todolist, rc.filter)
            if rc.stati == ["started"]:
                rc.stati = PROJECTUM_ACTIVE_STATI
            print(
                "If the indices are far from being in numerical order, please renumber them by running "
                "regolith helper u_todo -r"
            )
            print("Please choose from one of the following to update:")
            print("(index) action (days to due date|importance|expected duration (mins)|assigned by)")
            print("-" * 80)
            print_task(todolist, stati=rc.stati)
        else:
            match_todo = [i for i in todolist if i.get("running_index") == rc.index]
            if len(match_todo) == 0:
                raise RuntimeError("Please enter a valid index.")
            else:
                todo = match_todo[0]
                if rc.description:
                    todo["description"] = rc.description
                if rc.due_date:
                    try:
                        relative_day = int(rc.due_date)
                        due_date = today + relativedelta(days=relative_day)
                    except ValueError:
                        due_date = date_parser.parse(rc.due_date).date()
                    todo["due_date"] = due_date
                if rc.deadline:
                    if rc.deadline.lower() != "t":
                        if rc.deadline.lower() != "f":
                            raise RuntimeError("ERROR: allowed values for deadline are t or f")
                        else:
                            todo["deadline"] = False
                    else:
                        todo["deadline"] = True
                if rc.estimated_duration:
                    todo["duration"] = rc.estimated_duration
                if rc.importance or rc.importance == 0:
                    if rc.importance in ALLOWED_IMPORTANCE:
                        todo["importance"] = rc.importance
                    else:
                        raise ValueError(f"Importance should be chosen from{ALLOWED_IMPORTANCE}.")
                if rc.status:
                    if rc.status in TODO_STATI:
                        todo["status"] = rc.status
                    else:
                        raise ValueError(f"Status should be chosen from {TODO_STATI}.")
                if rc.notes:
                    try:
                        todo["notes"].extend(rc.notes)
                    except KeyError:
                        todo["notes"] = []
                        todo["notes"].extend(rc.notes)
                if rc.tags:
                    try:
                        todo["tags"].extend(rc.tags)
                    except KeyError:
                        todo["tags"] = []
                        todo["tags"].extend(rc.tags)
                if rc.begin_date:
                    todo["begin_date"] = date_parser.parse(rc.begin_date).date()
                if rc.end_date:
                    todo["end_date"] = date_parser.parse(rc.end_date).date()

                for i in range(0, len(rc.databases)):
                    db_name = rc.databases[i]["name"]
                    person_update = rc.client.find_one(db_name, rc.coll, filterid)
                    if person_update:
                        todolist_update = person_update.get("todos", [])
                    else:
                        continue
                    if len(todolist_update) != 0:
                        for i, todo_u in enumerate(todolist_update):
                            if rc.index == todo_u.get("running_index"):
                                todolist_update[i] = todo
                                rc.client.update_one(
                                    db_name,
                                    rc.coll,
                                    {"_id": rc.assigned_to},
                                    {"todos": todolist_update},
                                    upsert=True,
                                )
                                print(
                                    f"The task \"({todo_u['running_index']}) {todo_u['description'].strip()}\" in "
                                    f"{db_name} for {rc.assigned_to} has been updated."
                                )
                                return
        return



"""Contains a client database backed by the file system."""

import datetime
import json
import logging
import os
import signal
import sys
from collections import defaultdict
from copy import deepcopy
from glob import iglob

import ruamel.yaml
from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap, CommentedSeq

from regolith.tools import dbpathname


class DelayedKeyboardInterrupt:

    def __enter__(self):
        self.signal_received = False
        self.old_handler = signal.signal(signal.SIGINT, self.handler)

    def handler(self, sig, frame):
        self.signal_received = (sig, frame)
        logging.debug("SIGINT received. Delaying KeyboardInterrupt.")

    def __exit__(self, type, value, traceback):
        signal.signal(signal.SIGINT, self.old_handler)
        if self.signal_received:
            self.old_handler(*self.signal_received)


YAML_BASE_MAP = {CommentedMap: dict, CommentedSeq: list}


def _rec_re_type(i):
    """Destroy this when ruamel.yaml supports basetypes again"""
    if type(i) in YAML_BASE_MAP:
        base = YAML_BASE_MAP[type(i)]()
        if isinstance(base, dict):
            for k, v in i.items():
                base[_rec_re_type(k)] = _rec_re_type(v)
        elif isinstance(base, list):
            for j in i:
                base.append(_rec_re_type(j))
    else:
        base = i
    return base


def _id_key(doc):
    return doc["_id"]


def load_json(filename):
    """Loads a JSON file and returns a dict of its documents."""
    docs = {}
    with open(filename, encoding="utf-8") as fh:
        lines = fh.readlines()
    for line in lines:
        doc = json.loads(line)
        docs[doc["_id"]] = doc
    return docs


def date_encoder(obj):
    if isinstance(obj, (datetime.date, datetime.datetime)):
        return obj.isoformat()


def dump_json(filename, docs, date_handler=None):
    """Dumps a dict of documents into a file."""
    docs = sorted(docs.values(), key=_id_key)
    lines = [json.dumps(doc, sort_keys=True, default=date_handler) for doc in docs]
    s = "\n".join(lines)
    with open(filename, "w", encoding="utf-8") as fh:
        fh.write(s)


def load_yaml(filename, return_inst=False, loader=None):
    """Loads a YAML file and returns a dict of its documents."""
    if loader is None:
        inst = YAML()
    else:
        inst = loader
    with open(filename, encoding="utf-8") as fh:
        docs = inst.load(fh)
        docs = _rec_re_type(docs)
    for _id, doc in docs.items():
        doc["_id"] = _id
    return (docs, inst) if return_inst else docs


def dump_yaml(filename, docs, inst=None):
    """Dumps a dict of documents into a file."""
    inst = YAML() if inst is None else inst
    inst.representer.ignore_aliases = lambda *data: True
    inst.indent(mapping=2, sequence=4, offset=2)
    sorted_dict = ruamel.yaml.comments.CommentedMap()
    for k in sorted(docs):
        doc = docs[k]
        doc.pop("_id")
        sorted_dict[k] = ruamel.yaml.comments.CommentedMap()
        for kk in sorted(doc.keys()):
            sorted_dict[k][kk] = doc[kk]
    with open(filename, "w", encoding="utf-8") as fh:
        with DelayedKeyboardInterrupt():
            inst.dump(sorted_dict, stream=fh)


def json_to_yaml(inp, out):
    """Converts a JSON file to a YAML one."""
    docs = load_json(inp)
    dump_yaml(out, docs)


def yaml_to_json(inp, out, loader=None):
    """Converts a YAML file to a JSON one."""
    docs = load_yaml(inp, loader=loader)
    dump_json(out, docs)


class FileSystemClient:
    """A client database backed by the file system."""

    def __init__(self, rc):
        self.rc = rc
        self.closed = True
        self.dbs = None
        self.chained_db = None
        self.open()
        self._collfiletypes = {}
        self._collexts = {}
        self._yamlinsts = {}

    def is_alive(self):
        return not self.closed

    def open(self):
        if self.closed:
            self.dbs = defaultdict(lambda: defaultdict(dict))
            self.chained_db = {}
            self.closed = False

    def load_json(self, db, dbpath):
        """Loads the JSON part of a database."""
        dbs = self.dbs
        for f in [
            file
            for file in iglob(os.path.join(dbpath, "*.json"))
            if file not in db["blacklist"]
            and len(db["whitelist"]) == 0
            or os.path.basename(file).split(".")[0] in db["whitelist"]
        ]:
            collfilename = os.path.split(f)[-1]
            base, ext = os.path.splitext(collfilename)
            self._collfiletypes[base] = "json"
            print("loading " + f + "...", file=sys.stderr)
            dbs[db["name"]][base] = load_json(f)

    def load_yaml(self, db, dbpath):
        """Loads the YAML part of a database."""
        dbs = self.dbs
        for f in [
            file
            for file in iglob(os.path.join(dbpath, "*.y*ml"))
            if file not in db["blacklist"]
            and len(db["whitelist"]) == 0
            or os.path.basename(file).split(".")[0] in db["whitelist"]
        ]:
            collfilename = os.path.split(f)[-1]
            base, ext = os.path.splitext(collfilename)
            self._collexts[base] = ext
            self._collfiletypes[base] = "yaml"
            # print("loading " + f + "...", file=sys.stderr)
            coll, inst = load_yaml(f, return_inst=True)
            dbs[db["name"]][base] = coll
            self._yamlinsts[dbpath, base] = inst

    def load_database(self, db):
        """Loads a database."""
        dbpath = dbpathname(db, self.rc)
        self.load_json(db, dbpath)
        self.load_yaml(db, dbpath)

    def dump_json(self, docs, collname, dbpath):
        """Dumps json docs and returns filename"""
        f = os.path.join(dbpath, collname + ".json")
        dump_json(f, docs)
        filename = os.path.split(f)[-1]
        return filename

    def dump_yaml(self, docs, collname, dbpath):
        """Dumps json docs and returns filename"""
        f = os.path.join(dbpath, collname + self._collexts.get(collname, ".yaml"))
        inst = self._yamlinsts.get((dbpath, collname), None)
        dump_yaml(f, docs, inst=inst)
        filename = os.path.split(f)[-1]
        return filename

    def dump_database(self, db):
        """Dumps a database back to the filesystem."""
        dbpath = dbpathname(db, self.rc)
        os.makedirs(dbpath, exist_ok=True)
        to_add = []
        for collname, collection in self.dbs[db["name"]].items():
            # print("dumping " + collname + "...", file=sys.stderr)
            filetype = self._collfiletypes.get(collname, "yaml")
            if filetype == "json":
                filename = self.dump_json(collection, collname, dbpath)
            elif filetype == "yaml":
                filename = self.dump_yaml(collection, collname, dbpath)
            else:
                raise ValueError("did not recognize file type for regolith")
            to_add.append(os.path.join(db["path"], filename))
        return to_add

    def close(self):
        self.dbs = None
        self.closed = True

    def keys(self):
        return self.dbs.keys()

    def __getitem__(self, key):
        return self.dbs[key]

    def collection_names(self, dbname, include_system_collections=True):
        """Returns the collaction names for a database."""
        return set(self.dbs[dbname].keys())

    def all_documents(self, collname, copy=True):
        """Returns an iteratable over all documents in a collection."""
        if copy:
            return deepcopy(self.chained_db.get(collname, {})).values()
        return self.chained_db.get(collname, {}).values()

    def insert_one(self, dbname, collname, doc):
        """Inserts one document to a database/collection."""
        coll = self.dbs[dbname][collname]
        coll[doc["_id"]] = doc

    def insert_many(self, dbname, collname, docs):
        """Inserts many documents into a database/collection."""
        coll = self.dbs[dbname][collname]
        for doc in docs:
            coll[doc["_id"]] = doc

    def delete_one(self, dbname, collname, doc):
        """Removes a single document from a collection"""
        coll = self.dbs[dbname][collname]
        del coll[doc["_id"]]

    def find_one(self, dbname, collname, filter):
        """Finds the first document matching filter."""
        coll = self.dbs[dbname][collname]
        for doc in coll.values():
            matches = True
            for key, value in filter.items():
                if key not in doc or doc[key] != value:
                    matches = False
                    break
            if matches:
                return doc

    def update_one(self, dbname, collname, filter, update, **kwargs):
        """Updates one document."""
        coll = self.dbs[dbname][collname]
        doc = self.find_one(dbname, collname, filter)
        newdoc = dict(filter if doc is None else doc)
        newdoc.update(update)
        coll[newdoc["_id"]] = newdoc



"""A collection of python stylers"""

import re

month_fullnames = {
    1: "January",
    2: "February",
    3: "March",
    4: "April",
    5: "May",
    6: "June",
    7: "July",
    8: "August",
    9: "September",
    10: "October",
    11: "November",
    12: "December",
}
month_threelets = {
    1: "Jan",
    2: "Feb",
    3: "Mar",
    4: "Apr",
    5: "May",
    6: "Jun",
    7: "Jul",
    8: "Aug",
    9: "Sept",
    10: "Oct",
    11: "Nov",
    12: "Dec",
}


def sentencecase(sentence):
    """returns a sentence in sentencecase but with text in braces preserved

    Parameters
    ----------
    sentence: str
        The sentence

    Returns
    -------
    The sentence in sentence-case (but preserving any text  wrapped in braces)

    Notes
    -----
    tbd or n/a are returned lower case, not sentence case.
    """
    freezecaps = re.findall("\\{[^{}]+\\}", sentence)
    datalow = sentence.capitalize()
    sentencecase = re.split("\\{[^{}]+\\}", datalow)
    title = ""
    freezecaps.reverse()
    for word in sentencecase:
        if len(freezecaps) > 0:
            title = title + word + freezecaps.pop()
        else:
            title = title + word
    if title.lower() == "n/a":
        title = title.lower()
    if title.lower() == "tbd":
        title = title.lower()
    return title



"""The main CLI for regolith"""

from __future__ import print_function

import copy
import os

from gooey import Gooey, GooeyParser

from regolith import commands
from regolith.commands import CONNECTED_COMMANDS
from regolith.database import connect
from regolith.helper import HELPERS
from regolith.runcontrol import DEFAULT_RC, filter_databases, load_rcfile
from regolith.schemas import SCHEMAS
from regolith.tools import update_schemas

NEED_RC = set(CONNECTED_COMMANDS.keys())
NEED_RC |= {"rc", "deploy", "store"}


# @Gooey(advanced=True)
@Gooey(  # body_bg_color='#808080',
    # header_bg_color='#808080',
    required_cols=1,
    optional_cols=1,
    sidebar_title="Helpers",
    program_name="Regolith Helper GUI",
)
def create_parser():
    p = GooeyParser()
    subp = p.add_subparsers(title="helper_target", dest="helper_target")
    for k, v in HELPERS.items():
        subpi = subp.add_parser(k)
        v[1](subpi)

    return p


def main(args=None):
    rc = DEFAULT_RC
    parser = create_parser()
    ns = parser.parse_args()
    ns.cmd = "helper"
    if os.path.exists(rc.user_config):
        rc._update(load_rcfile(rc.user_config))
    rc._update(load_rcfile("regolithrc.json"))
    rc._update(ns.__dict__)
    if "schemas" in rc._dict:
        user_schema = copy.deepcopy(rc.schemas)
        default_schema = copy.deepcopy(SCHEMAS)
        rc.schemas = update_schemas(default_schema, user_schema)
    else:
        rc.schemas = SCHEMAS
    filter_databases(rc)
    dbs = commands.helper_db_check(rc)
    with connect(rc, dbs=dbs) as rc.client:
        CONNECTED_COMMANDS[rc.cmd](rc)


if __name__ == "__main__":
    main()



import datetime
import tempfile
from pathlib import Path

from regolith.fsclient import date_encoder, dump_json


def test_date_encoder():
    day = datetime.date(2021, 1, 1)
    time = datetime.datetime(2021, 5, 18, 6, 28, 21, 504549)
    assert date_encoder(day) == "2021-01-01"
    assert date_encoder(time) == "2021-05-18T06:28:21.504549"


def test_dump_json():
    doc = {
        "first": {"_id": "first", "name": "me", "date": datetime.date(2021, 5, 1), "test_list": [5, 4]},
        "second": {"_id": "second"},
    }
    json_doc = '{"_id": "first", "date": "2021-05-01", "name": "me", "test_list": [5, 4]}\n{"_id": "second"}'
    temp_dir = Path(tempfile.gettempdir())
    filename = temp_dir / "test.json"
    dump_json(filename, doc, date_handler=date_encoder)
    with open(filename, "r", encoding="utf-8") as f:
        actual = f.read()
    assert actual == json_doc


# datasets = [
#     (
#         {"first": {"date": "2021-05-01", "name": "me", "test_list": [5, 4]}, "second": {}},
#         {
#             "first": {"_id": "first", "name": "me", "date": datetime.date(2021, 5, 1), "test_list": [5, 4]},
#             "second": {"_id": "second"},
#         },
#     ),
# ]
# @pytest.mark.parametrize("json_doc, expected", datasets)
# def test_load_json(json_doc, expected):
#     temp_dir = Path(tempfile.gettempdir())
#     filename = temp_dir / "test.json"
#     with open(filename, "w", encoding="utf-8") as f:
#         json.dump(json_doc, f)
#     actual = load_json(filename)
#     assert actual == expected
